# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './component_widget_ui.ui',
# licensing of './component_widget_ui.ui' applies.
#
# Created: Sat Jun 19 22:02:29 2021
#      by: pyside6-uic  running on PySide2 5.13.2
#
# WARNING! All changes made in this file will be lost!

from PySide6 import QtCore, QtGui, QtWidgets


class Ui_ComponentWidget(object):

    def setupUi(self, ComponentWidget):
        ComponentWidget.setObjectName("ComponentWidget")
        ComponentWidget.resize(539, 475)
        ComponentWidget.setTabPosition(QtWidgets.QTabWidget.West)
        ComponentWidget.setTabShape(QtWidgets.QTabWidget.Triangular)
        ComponentWidget.setMovable(True)
        self.tabOptions = QtWidgets.QWidget()
        self.tabOptions.setObjectName("tabOptions")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.tabOptions)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setContentsMargins(-1, -1, -1, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.treeView = QTreeView_Options(self.tabOptions)
        self.treeView.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.treeView.setFrameShadow(QtWidgets.QFrame.Plain)
        self.treeView.setSizeAdjustPolicy(
            QtWidgets.QAbstractScrollArea.AdjustToContents)
        self.treeView.setProperty("showDropIndicator", False)
        self.treeView.setObjectName("treeView")
        self.verticalLayout.addWidget(self.treeView)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/options"), QtGui.QIcon.Normal,
                       QtGui.QIcon.On)
        ComponentWidget.addTab(self.tabOptions, icon, "")
        self.tabHelp = QtWidgets.QWidget()
        self.tabHelp.setObjectName("tabHelp")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.tabHelp)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.textHelp = QtWidgets.QTextEdit(self.tabHelp)
        self.textHelp.setObjectName("textHelp")
        self.verticalLayout_2.addWidget(self.textHelp)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/help"), QtGui.QIcon.Normal,
                        QtGui.QIcon.On)
        ComponentWidget.addTab(self.tabHelp, icon1, "")
        self.tabSource = QtWidgets.QWidget()
        self.tabSource.setObjectName("tabSource")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.tabSource)
        self.verticalLayout_3.setSpacing(0)
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.textSource = QtWidgets.QTextEdit(self.tabSource)
        self.textSource.setAutoFillBackground(True)
        self.textSource.setLineWrapMode(QtWidgets.QTextEdit.NoWrap)
        self.textSource.setObjectName("textSource")
        self.verticalLayout_3.addWidget(self.textSource)
        self.lineSourcePath = QtWidgets.QLineEdit(self.tabSource)
        self.lineSourcePath.setReadOnly(True)
        self.lineSourcePath.setClearButtonEnabled(False)
        self.lineSourcePath.setObjectName("lineSourcePath")
        self.verticalLayout_3.addWidget(self.lineSourcePath)
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/_imgs/support.png"),
                        QtGui.QIcon.Normal, QtGui.QIcon.On)
        ComponentWidget.addTab(self.tabSource, icon2, "")

        self.retranslateUi(ComponentWidget)
        ComponentWidget.setCurrentIndex(2)

    def retranslateUi(self, ComponentWidget):
        ComponentWidget.setWindowTitle(
            QtWidgets.QApplication.translate("ComponentWidget",
                                             "Edit a component", None, -1))
        ComponentWidget.setTabText(
            ComponentWidget.indexOf(self.tabOptions),
            QtWidgets.QApplication.translate("ComponentWidget", "Options", None,
                                             -1))
        ComponentWidget.setTabToolTip(
            ComponentWidget.indexOf(self.tabOptions),
            QtWidgets.QApplication.translate("ComponentWidget",
                                             "Edit the make options", None, -1))
        self.textHelp.setHtml(
            QtWidgets.QApplication.translate(
                "ComponentWidget",
                "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
                "<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
                "p, li { white-space: pre-wrap; }\n"
                "</style></head><body style=\" font-family:\'.AppleSystemUIFont\'; font-size:13pt; font-weight:400; font-style:normal;\">\n"
                "<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-family:\'Arial\'; font-size:14pt;\">Help about a component will be displayed here when you select a component in the design components dialog. </span></p></body></html>",
                None, -1))
        ComponentWidget.setTabText(
            ComponentWidget.indexOf(self.tabHelp),
            QtWidgets.QApplication.translate("ComponentWidget", "Help", None,
                                             -1))
        self.textSource.setHtml(
            QtWidgets.QApplication.translate(
                "ComponentWidget",
                "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
                "<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
                "p, li { white-space: pre-wrap; }\n"
                "</style></head><body style=\" font-family:\'.AppleSystemUIFont\'; font-size:13pt; font-weight:400; font-style:normal;\">\n"
                "<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-family:\'Arial\'; font-size:14pt;\">The source code of a QComponent class will be displayed here when you select a component in the design components dialog. </span></p></body></html>",
                None, -1))
        self.lineSourcePath.setToolTip(
            QtWidgets.QApplication.translate("ComponentWidget",
                                             "Source code file path", None, -1))
        self.lineSourcePath.setStatusTip(
            QtWidgets.QApplication.translate("ComponentWidget",
                                             "Source code file path", None, -1))
        self.lineSourcePath.setWhatsThis(
            QtWidgets.QApplication.translate("ComponentWidget",
                                             "Source code file path", None, -1))
        self.lineSourcePath.setAccessibleName(
            QtWidgets.QApplication.translate("ComponentWidget",
                                             "Source code file path", None, -1))
        self.lineSourcePath.setAccessibleDescription(
            QtWidgets.QApplication.translate("ComponentWidget",
                                             "Source code file path", None, -1))
        self.lineSourcePath.setText(
            QtWidgets.QApplication.translate("ComponentWidget",
                                             "Source code file path here", None,
                                             -1))
        ComponentWidget.setTabText(
            ComponentWidget.indexOf(self.tabSource),
            QtWidgets.QApplication.translate("ComponentWidget", "Source", None,
                                             -1))


from .widgets.edit_component.tree_view_options import QTreeView_Options
from . import main_window_rc_rc



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import numpy as np
from numpy.linalg import norm
from qiskit_metal import Dict
from qiskit_metal.qlibrary.core import QRoute
from qiskit_metal.toolbox_metal import math_and_overrides as mao


class RouteFramed(QRoute):
    """A non-meandered sample_shapes CPW that is auto-generated between 2 components.
    Designed to avoid self-collisions and collisions with components it is
    attached to.

    This class extends the `QComponent` class.

    .. meta::
        :description: Routed Frame

    Assumptions:
        1. Components are situated along xy axes in 2 dimensions. No rotation is allowed (yet). Their bounding boxes may
           not overlap, though they may be situated at arbitrary x and y provided these conditions are met.
        2. Pins point normal to qubits ("directly outward") and either in the x or y directions. They must not protrude
           from the exact corner of a component. [This last assumption has implications for 2-segment connections.]
        3. Intersection of CPWs with themselves or the qubits they stem from is prohibited. Intersection with other
           components/CPWs has not yet been considered.
        4. Components may not share an edge; a nonzero gap must be present between 2 adjacent qubits.
        5. CPWs must be attached to protruding leads via terminations head-on, not from the sides.
    """

    component_metadata = Dict(short_name='cpw')
    """Component metadata"""

    TOOLTIP = """A non-meandered basic CPW that is auto-generated between 2 components."""

    def make(self):
        """Use user-specified parameters and geometric orientation of
        components to determine whether the CPW connecting the pins on either
        end requires 1, 2, or 3 segments. Preference given to shorter paths and
        paths that flow with the leadin and leadout directions rather than
        turning immediately at 90 deg.

        Keepout region along x and y directions specified for CPWs that
        wrap around outer perimeter of overall bounding box of both
        components.
        """

        self.__pts = []  # list of 2D numpy arrays containing vertex locations

        p = self.p  # parsed options

        w = p.trace_width
        leadstart = p.lead.start_straight
        leadend = p.lead.end_straight
        keepoutx = 0.2
        keepouty = 0.2

        # Set the CPW pins and add the points/directions to the lead-in/out arrays
        self.set_pin("start")
        self.set_pin("end")

        # Align the lead-in/out to the input options set from the user
        meander_start_point = self.set_lead("start")
        meander_end_point = self.set_lead("end")

        n1 = meander_start_point.direction
        n2 = meander_end_point.direction

        m1 = meander_start_point.position
        m2 = meander_end_point.position

        component_start = p.pin_inputs['start_pin']['component']
        component_end = p.pin_inputs['end_pin']['component']

        # Coordinates of bounding box for each individual component
        minx1, miny1, maxx1, maxy1 = self.design.components[
            component_start].qgeometry_bounds()
        minx2, miny2, maxx2, maxy2 = self.design.components[
            component_end].qgeometry_bounds()

        # Coordinates of overall bounding box which includes both components
        minx, miny = min(minx1, minx2), min(miny1, miny2)
        maxx, maxy = max(maxx1, maxx2), max(maxy1, maxy2)

        # Normdot is dot product of normal vectors
        normdot = mao.dot(n1, n2)

        if normdot == -1:
            # Modify CPW endpoints to include mandatory w / 2 leadin plus user defined leadin
            m1ext = m1 + n1 * (w / 2 + leadstart)
            m2ext = m2 + n2 * (w / 2 + leadend)
            # Alignment is between displacement of modified positions (see above) and normal vector
            alignment = mao.dot((m2ext - m1ext) / norm(m2ext - m1ext), n1)
            if alignment == 1:
                # Normal vectors point directly at each other; no obstacles in between
                # Connect with single segment; generalizes to arbitrary angles with no snapping
                self.__pts = self.connect_frame(1)
            elif alignment > 0:
                # Displacement partially aligned with normal vectors
                # Connect with antisymmetric 3-segment CPW
                if n1[1] == 0:
                    # Normal vectors horizontal
                    if minx2 < maxx1:
                        self.__pts = self.connect_frame(3, 0,
                                                        (minx2 + maxx1) / 2)
                    else:
                        self.__pts = self.connect_frame(3, 0,
                                                        (minx1 + maxx2) / 2)
                else:
                    # Normal vectors vertical
                    if miny2 < maxy1:
                        self.__pts = self.connect_frame(3, 1,
                                                        (miny2 + maxy1) / 2)
                    else:
                        self.__pts = self.connect_frame(3, 1,
                                                        (miny1 + maxy2) / 2)
            elif alignment == 0:
                # Displacement orthogonal to normal vectors
                # Connect with 1 segment
                self.__pts = self.connect_frame(1)
            elif alignment < 0:
                # Normal vectors on opposite sides of squares
                if n1[1] == 0:
                    # Normal vectors horizontal
                    if maxy1 < miny2:
                        self.__pts = self.connect_frame(3, 1,
                                                        (maxy1 + miny2) / 2)
                    elif maxy2 < miny1:
                        self.__pts = self.connect_frame(3, 1,
                                                        (maxy2 + miny1) / 2)
                    else:
                        # Gap running only vertically -> must wrap around with shorter of 2 possibilities
                        # pts_top represents 3-segment CPW running along top edge of overall bounding box
                        # pts_bott represents 3-segment CPW running along bottom edge of overall bounding box
                        pts_top = self.connect_frame(3, 1, maxy + keepouty)
                        pts_bott = self.connect_frame(3, 1, miny - keepouty)
                        if self.totlength(pts_top) < self.totlength(pts_bott):
                            self.__pts = pts_top
                        else:
                            self.__pts = pts_bott
                else:
                    # Normal vectors vertical
                    if maxx1 < minx2:
                        self.__pts = self.connect_frame(3, 0,
                                                        (maxx1 + minx2) / 2)
                    elif maxx2 < minx1:
                        self.__pts = self.connect_frame(3, 0,
                                                        (maxx2 + minx1) / 2)
                    else:
                        # Gap running only horizontally -> must wrap around with shorter of 2 possibilities
                        # pts_left represents 3-segment CPW running along left edge of overall bounding box
                        # pts_right represents 3-segment CPW running along right edge of overall bounding box
                        pts_left = self.connect_frame(3, 0, minx - keepoutx)
                        pts_right = self.connect_frame(3, 0, maxx + keepoutx)
                        if self.totlength(pts_left) < self.totlength(pts_right):
                            self.__pts = pts_left
                        else:
                            self.__pts = pts_right
        elif normdot == 0:
            # Normal vectors perpendicular to each other
            if (m1[0] in [minx, maxx]) and (m2[1] in [miny, maxy]):
                # Both pins on perimeter of overall bounding box, but not at corner
                self.__pts = self.connect_frame(2)
            elif (m1[1] in [miny, maxy]) and (m2[0] in [minx, maxx]):
                # Both pins on perimeter of overall bounding box, but not at corner
                self.__pts = self.connect_frame(2)
            elif (m1[0] not in [minx, maxx]) and (m2[0] not in [
                    minx, maxx
            ]) and (m1[1] not in [miny, maxy]) and (m2[1] not in [miny, maxy]):
                # Neither pin lies on perimeter of overall bounding box
                # Always possible to connect with at most 2 segments
                if (m1[0] == m2[0]) or (m1[1] == m2[1]):
                    # Connect directly with 1 segment
                    self.__pts = self.connect_frame(1)
                else:
                    # Connect with 2 segments
                    self.__pts = self.connect_frame(2)
            elif (m1[0] in [minx, maxx]) or (m1[1] in [miny, maxy]):
                # Pin 1 lies on boundary of overall bounding box but pin 2 does not
                if m1[0] in [minx, maxx]:
                    # Pin 1 on left or right boundary, pointing left or right, respectively
                    if n2[1] > 0:
                        # Pin 2 pointing up
                        if miny1 > maxy2:
                            self.__pts = self.connect_frame(2)
                        else:
                            self.__pts = self.connect_frame(
                                3, 1, maxy + keepouty)
                    else:
                        # Pin 2 pointing down
                        if miny2 > maxy1:
                            self.__pts = self.connect_frame(2)
                        else:
                            self.__pts = self.connect_frame(
                                3, 1, miny - keepouty)
                elif m1[1] in [miny, maxy]:
                    # Pin 1 on bottom or top boundary, pointing down or up, respectively
                    if n2[0] < 0:
                        # Pin 2 pointing left
                        if minx2 > maxx1:
                            self.__pts = self.connect_frame(2)
                        else:
                            self.__pts = self.connect_frame(
                                3, 0, minx - keepoutx)
                    else:
                        # Pin 2 pointing right
                        if minx1 > maxx2:
                            self.__pts = self.connect_frame(2)
                        else:
                            self.__pts = self.connect_frame(
                                3, 0, maxx + keepoutx)
            elif (m2[0] in [minx, maxx]) or (m2[1] in [miny, maxy]):
                # Pin 2 lies on boundary of overall bounding box but pin 1 does not
                if m2[0] in [minx, maxx]:
                    # Pin 2 on left or right boundary, pointing left or right, respectively
                    if n1[1] > 0:
                        # Pin 1 pointing up
                        if miny2 > maxy1:
                            self.__pts = self.connect_frame(2)
                        else:
                            self.__pts = self.connect_frame(
                                3, 1, maxy + keepouty)
                    else:
                        # Pin 1 pointing down
                        if miny1 > maxy2:
                            self.__pts = self.connect_frame(2)
                        else:
                            self.__pts = self.connect_frame(
                                3, 1, miny - keepouty)
                elif m2[1] in [miny, maxy]:
                    # Pin 2 on bottom or top boundary, pointing down or up, respectively
                    if n1[0] < 0:
                        # Pin 1 pointing left
                        if minx1 > maxx2:
                            self.__pts = self.connect_frame(2)
                        else:
                            self.__pts = self.connect_frame(
                                3, 0, minx - keepoutx)
                    else:
                        # Pin 1 pointing right
                        if minx2 > maxx1:
                            self.__pts = self.connect_frame(2)
                        else:
                            self.__pts = self.connect_frame(
                                3, 0, maxx + keepoutx)
        else:
            # Normal vectors pointing in same direction
            if ((m1[0] == m2[0]) or
                (m1[1] == m2[1])) and (mao.dot(n1, m2 - m1) == 0):
                # Connect directly with 1 segment
                self.__pts = self.connect_frame(1)
            elif n1[1] == 0:
                # Normal vectors horizontal
                if (m1[1] > maxy2) or (m2[1] > maxy1):
                    # Connect with 2 segments
                    self.__pts = self.connect_frame(2)
                else:
                    # Must wrap around with shorter of the 2 following possibilities:
                    # pts_top represents 3-segment CPW running along top edge of overall bounding box
                    # pts_bott represents 3-segment CPW running along bottom edge of overall bounding box
                    pts_top = self.connect_frame(3, 1, maxy + keepouty)
                    pts_bott = self.connect_frame(3, 1, miny - keepouty)
                    if self.totlength(pts_top) < self.totlength(pts_bott):
                        self.__pts = pts_top
                    else:
                        self.__pts = pts_bott
            else:
                # Normal vectors vertical
                if (m1[0] > maxx2) or (m2[0] > maxx1):
                    # Connect with 2 segments
                    self.__pts = self.connect_frame(2)
                else:
                    # Must wrap around with shorter of the 2 following possibilities:
                    # pts_left represents 3-segment CPW running along left edge of overall bounding box
                    # pts_right represents 3-segment CPW running along right edge of overall bounding box
                    pts_left = self.connect_frame(3, 0, minx - keepoutx)
                    pts_right = self.connect_frame(3, 0, maxx + keepoutx)
                    if self.totlength(pts_left) < self.totlength(pts_right):
                        self.__pts = pts_left
                    else:
                        self.__pts = pts_right

        self.intermediate_pts = self.__pts

        # Make points into elements
        self.make_elements(self.get_points())

    def connect_frame(self, segments: int, constaxis=0, constval=0) -> list:
        """Generate the list of 2D coordinates comprising a CPW between
        startpin and endpin.

        Args:
            startpin (str): Name of startpin
            endpin (str): Name of endpin
            width (float): Width of CPW in mm
            segments (int): Number of segments in the CPW, not including leadin and leadout. Ranges from 1 to 3.
            leadstart (float): Length of first CPW segment originating at startpin
            leadend (float): Length of final CPW segment ending at endpin
            constaxis (int, optional): In the case of 3 segment CPWs, the constant axis of the line that both
                leadin and leadout must connect to. Example: If x = 3, the constant axis (x) is 0.  Defaults to 0.
            constval (int, optional): In the case of 3 segment CPWs, the constant numerical value of the line
                that both leadin and leadout must connect to. Example: If x = 3, the constant value is 3.  Defaults to 0.

        Returns:
            List: [np.array([x0, y0]), np.array([x1, y1]), np.array([x2, y2])] where xi, yi are vertices of CPW.
        """

        start = self.head.get_tip()
        end = self.tail.get_tip()

        if segments == 1:
            # Straight across or up and down; no additional points necessary
            midcoords = []
        elif segments == 2:
            # Choose between 2 diagonally opposing corners so that CPW doesn't trace back on itself
            corner1 = np.array([(start.position)[0], (end.position)[1]])
            corner2 = np.array([(end.position)[0], (start.position)[1]])
            startc1 = mao.dot(corner1 - (start.position), start.direction)
            endc1 = mao.dot(corner1 - (end.position), end.direction)
            startc2 = mao.dot(corner2 - (start.position), start.direction)
            endc2 = mao.dot(corner2 - (end.position), end.direction)
            # If both pins' normal vectors point towards one of the corners, pick that corner
            if (startc1 > 0) and (endc1 > 0):
                midcoords = [corner1]
            elif (startc2 > 0) and (endc2 > 0):
                midcoords = [corner2]
            elif min(startc1, endc1) >= 0:
                midcoords = [corner1]
            else:
                midcoords = [corner2]
        elif segments == 3:
            # Connect start and end to long segment at constaxis
            # 2 additional intermediate points needed to achieve this
            if constaxis == 0:
                # Middle segment lies on the line x = c for some constant c
                midcoord_start = np.array([constval, (start.position)[1]])
                midcoord_end = np.array([constval, (end.position)[1]])
            else:
                # Middle segment lies on the line y = d for some constant d
                midcoord_start = np.array([(start.position)[0], constval])
                midcoord_end = np.array([(end.position)[0], constval])
            midcoords = [midcoord_start, midcoord_end]
        startpts = [start.position]
        endpts = [end.position]
        return startpts + midcoords + endpts

    def totlength(self, pts: list) -> float:
        """Get total length of all line segments in a given CPW."""
        return sum(norm(pts[i] - pts[i - 1]) for i in range(1, len(pts)))



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# modified by Chalmers/SK/20210621
# modified by Samarth Hawaldar

from collections import defaultdict
from pathlib import Path
from typing import Union, Tuple

import logging
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pyEPR as epr
from pyEPR.ansys import set_property, parse_units

from qiskit_metal import Dict
from qiskit_metal.draw.utility import to_vec3D
from qiskit_metal.renderers.renderer_ansys.ansys_renderer import (
    QAnsysRenderer, get_clean_name)


class QHFSSRenderer(QAnsysRenderer):
    """Subclass of QAnsysRenderer for running HFSS simulations.

    QAnsysRenderer Default Options:
        * Lj: '10nH' -- Lj has units of nanoHenries (nH)
        * Cj: 0 -- Cj *must* be 0 for pyEPR analysis! Cj has units of femtofarads (fF)
        * _Rj: 0 -- _Rj *must* be 0 for pyEPR analysis! _Rj has units of Ohms
        * max_mesh_length_jj: '7um' -- Maximum mesh length for Josephson junction elements
        * max_mesh_length_port: '7um' -- Maximum mesh length for Ports in Eigenmode Simulations
        * project_path: None -- Default project path; if None --> get active
        * project_name: None -- Default project name
        * design_name: None -- Default design name
        * ansys_file_extension: '.aedt' -- Ansys file extension for 2016 version and newer
        * x_buffer_width_mm: 0.2 -- Buffer between max/min x and edge of ground plane, in mm
        * y_buffer_width_mm: 0.2 -- Buffer between max/min y and edge of ground plane, in mm
    """

    name = 'hfss'
    """Name"""

    hfss_options = Dict(
        port_inductor_gap=
        '10um'  # spacing between port and inductor if junction is drawn both ways
    )
    """HFSS Options"""

    def __init__(self, design: 'QDesign', initiate=True, options: Dict = None):
        """Create a QRenderer for HFSS simulations, subclassed from QAnsysRenderer.

        Args:
            design (QDesign): Use QGeometry within QDesign to obtain elements for Ansys.
            initiate (bool, optional): True to initiate the renderer. Defaults to True.
            options (Dict, optional):  Used to override all options. Defaults to None.
        """
        super().__init__(design=design, initiate=initiate, options=options)

        self.current_sweep = None

        QHFSSRenderer.load()

    def render_design(self,
                      selection: Union[list, None] = None,
                      open_pins: Union[list, None] = None,
                      port_list: Union[list, None] = None,
                      jj_to_port: Union[list, None] = None,
                      ignored_jjs: Union[list, None] = None,
                      box_plus_buffer: bool = True):
        """Initiate rendering of components in design contained in selection,
        assuming they're valid. Components are rendered before the chips they
        reside on, and subtraction of negative shapes is performed at the very
        end. Add the metalize() method here to turn objects in
        self.assign_perfE (see init in QAnsysRenderer class) into perfect
        electrical conductors. Create lumped ports as needed.

        First obtain a list of IDs of components to render and a corresponding
        case, denoted by self.qcomp_ids and self.case, respectively. If
        self.case == 1, all components in QDesign are to be rendered.  If
        self.case == 0, a strict subset of components in QDesign are to be
        rendered. Otherwise, if self.case == 2, one or more component names in
        selection cannot be found in QDesign.

        Among the components selected for export, there may or may not be
        unused (unconnected) pins. The second parameter, open_pins, contains
        tuples of the form (component_name, pin_name) that specify exactly
        which pins should be open rather than shorted during the simulation.
        Both the component and pin name must be specified because the latter
        could be shared by multiple components. All pins in this list are
        rendered with an additional endcap in the form of a rectangular
        cutout, to be subtracted from its respective plane.

        In driven modal solutions, the Ansys design must include one or more
        ports. This is done by adding all port designations and their
        respective impedances in Ohms as (qcomp, pin, impedance) to port_list.
        Note that an open endcap must separate the two sides of each pin before
        inserting a lumped port in between, so behind the scenes all pins in
        port_list are also added to open_pins. Practically, however, port_list
        and open_pins are inputted as mutually exclusive lists.

        Also in driven modal solutions, one may want to render junctions as
        lumped ports and/or inductors, or omit them altogether. To do so,
        tuples of the form (component_name, element_name, impedance, draw_ind)
        are added to the list jj_to_port. For example,
        ('Q1', 'rect_jj', 70, True) indicates that rect_jj of component Q1 is
        to be rendered as both a lumped port with an impedance of 70 Ohms as
        well as an inductor whose properties are given in default_options.
        Replacing the last entry of this 4-element tuple with False would
        indicate that only the port is to be drawn, not the inductor.
        Alternatively for driven modal solutions, one may want to disregard
        select junctions in the Metal design altogether to simulate the
        capacitive effect while keeping the qubit in an "off" state. Such
        junctions are specified in the form (component_name, element_name)
        in the list ignored_jjs.

        The final parameter, box_plus_buffer, determines how the chip is drawn.
        When set to True, it takes the minimum rectangular bounding box of all
        rendered components and adds a buffer of x_buffer_width_mm and
        y_buffer_width_mm horizontally and vertically, respectively, to the
        chip size. The center of the chip lies at the midpoint x/y coordinates
        of the minimum rectangular bounding box and may change depending on
        which components are rendered and how they're positioned. If
        box_plus_buffer is False, however, the chip position and dimensions
        are taken from the chip info dictionary found in self.design,
        irrespective of what's being rendered. While this latter option is
        faster because it doesn't require calculating a bounding box, it runs
        the risk of rendered components being too close to the edge of the chip
        or even falling outside its boundaries.

        Args:
            selection (Union[list, None], optional): List of components to
                                        render. Defaults to None.
            open_pins (Union[list, None], optional): List of tuples of pins
                                        that are open. Defaults to None.
            port_list (Union[list, None], optional): List of tuples of pins to
                                        be rendered as ports. Defaults to None.
            jj_to_port (Union[list, None], optional): List of tuples of jj's to
                                        be rendered as ports. Defaults to None.
            ignored_jjs (Union[list, None], optional): List of tuples of jj's
                                        that shouldn't be rendered.
                                        Defaults to None.
            box_plus_buffer (bool): Either calculate a bounding box based on
                                        the location of rendered geometries
                                        or use chip size from design class.
        """
        self.qcomp_ids, self.case = self.get_unique_component_ids(selection)

        if self.case == 2:
            self.logger.warning(
                'Unable to proceed with rendering. Please check selection.')
            return

        self.chip_subtract_dict = defaultdict(set)
        self.assign_perfE = []
        self.assign_mesh = []
        self.assign_port_mesh = []
        self.jj_lumped_ports = {}
        self.jj_to_ignore = set()

        if jj_to_port:
            self.jj_lumped_ports = {
                (qcomp, elt): [impedance, draw_ind]
                for qcomp, elt, impedance, draw_ind in jj_to_port
            }

        if ignored_jjs:
            self.jj_to_ignore = {(qcomp, qelt) for qcomp, qelt in ignored_jjs}

        self.render_tables()
        if port_list:
            self.add_endcaps(open_pins +
                             [(qcomp, pin) for qcomp, pin, _ in port_list])
        else:
            self.add_endcaps(open_pins)

        self.render_chips(box_plus_buffer=box_plus_buffer)
        self.subtract_from_ground()
        if port_list:
            self.create_ports(port_list)
        self.add_mesh()
        self.metallize()

    def create_ports(self, port_list: list):
        """Add ports and their respective impedances in Ohms to designated pins
        in port_list. Port_list is formatted as [(qcomp_0, pin_0, impedance_0),
        (qcomp_1, pin_1, impedance_1), ...].

        Args:
            port_list (list): List of tuples of pins to be rendered as ports.
        """
        for qcomp, pin, impedance in port_list:
            port_name = f'Port_{qcomp}_{pin}'
            pdict = self.design.components[qcomp].pins[pin]
            midpt, gap_size, norm_vec, width = pdict['middle'], pdict['gap'], \
                                               pdict['normal'], pdict['width']
            width = parse_units(width)
            endpoints = parse_units([midpt, midpt + gap_size * norm_vec])
            endpoints_3d = to_vec3D(endpoints, 0)  # Set z height to 0
            x0, y0 = endpoints_3d[0][:2]
            x1, y1 = endpoints_3d[1][:2]
            if abs(y1 - y0) > abs(x1 - x0):
                # Junction runs vertically up/down
                x_min, x_max = x0 - width / 2, x0 + width / 2
                y_min, y_max = min(y0, y1), max(y0, y1)
            else:
                # Junction runs horizontally left/right
                x_min, x_max = min(x0, x1), max(x0, x1)
                y_min, y_max = y0 - width / 2, y0 + width / 2

            # Draw rectangle
            self.logger.debug(f'Drawing a rectangle: {port_name}')
            poly_ansys = self.modeler.draw_rect_corner([x_min, y_min, 0],
                                                       x_max - x_min,
                                                       y_max - y_min, 0,
                                                       **dict(transparency=0.0))
            axis = 'x' if abs(x1 - x0) > abs(y1 - y0) else 'y'
            if self.pinfo.design.solution_type != 'Eigenmode':
                poly_ansys.make_lumped_port(axis,
                                            z0=str(impedance) + 'ohm',
                                            name=f'LumpPort_{qcomp}_{pin}')
                self.modeler.rename_obj(poly_ansys, port_name)
            else:
                poly_ansys.make_rlc_boundary(axis,
                                             r=str(impedance) + 'ohm',
                                             name=f'RLCBoundary_{qcomp}_{pin}')
                self.modeler.rename_obj(poly_ansys, port_name)
                self.assign_port_mesh.append(port_name)

            # Draw line
            lump_line = self.modeler.draw_polyline(
                [endpoints_3d[0], endpoints_3d[1]],
                closed=False,
                **dict(color=(128, 0, 128)))
            lump_line = lump_line.rename(f'voltage_line_{port_name}')
            lump_line.show_direction = True

    def render_element_junction(self, qgeom: pd.Series):
        """
        Render a Josephson junction depending on the solution type.

        If in HFSS eigenmode, junctions are rendered as inductors consisting of
            1. A rectangle of length pad_gap and width inductor_width. Defines lumped element RLC
               boundary condition.
            2. A line that is later used to calculate the voltage in post-processing analysis.

        If in HFSS driven modal, junctions can be inductors, lumped ports, both inductors
        and lumped ports, or omitted altogether. Ports are characterized by an impedance
        value given in the list jj_to_port when render_design() is called.

        Args:
            qgeom (pd.Series): GeoSeries of element properties.
        """
        qcomp = self.design._components[qgeom['component']].name
        qc_elt = get_clean_name(qgeom['name'])

        if (qcomp, qc_elt) not in self.jj_to_ignore:
            qc_shapely = qgeom.geometry
            qc_chip_z = parse_units(self.design.get_chip_z(qgeom.chip))
            qc_width = parse_units(qgeom.width)

            endpoints = parse_units(list(qc_shapely.coords))
            endpoints_3d = to_vec3D(endpoints, qc_chip_z)
            x0, y0, z0 = endpoints_3d[0]
            x1, y1, z0 = endpoints_3d[1]
            if abs(y1 - y0) > abs(x1 - x0):
                # Junction runs vertically up/down
                axis = 'y'
                x_min, x_max = x0 - qc_width / 2, x0 + qc_width / 2
                y_min, y_max = min(y0, y1), max(y0, y1)
            else:
                # Junction runs horizontally left/right
                axis = 'x'
                x_min, x_max = min(x0, x1), max(x0, x1)
                y_min, y_max = y0 - qc_width / 2, y0 + qc_width / 2

            if (qcomp, qc_elt) in self.jj_lumped_ports:
                if self.jj_lumped_ports[(qcomp, qc_elt)][1]:
                    # Draw both port and inductor side by side with small gap in between
                    gap = parse_units(self.hfss_options['port_inductor_gap'])
                    x_mid, y_mid = (x_min + x_max) / 2, (y_min + y_max) / 2
                    if axis == 'x':
                        y_mid_hi = y_mid + gap / 2
                        y_mid_lo = y_mid - gap / 2
                        self.render_junction_port(qgeom, x_min, x_max, y_mid_hi,
                                                  y_max, qc_chip_z, axis)
                        self.render_junction_inductor(qgeom, x_min, x_max,
                                                      y_min, y_mid_lo,
                                                      qc_chip_z, axis)
                    elif axis == 'y':
                        x_mid_lo = x_mid - gap / 2
                        x_mid_hi = x_mid + gap / 2
                        self.render_junction_port(qgeom, x_mid_hi, x_max, y_min,
                                                  y_max, qc_chip_z, axis)
                        self.render_junction_inductor(qgeom, x_min, x_mid_lo,
                                                      y_min, y_max, qc_chip_z,
                                                      axis)
                else:
                    # Only draw port
                    self.render_junction_port(qgeom, x_min, x_max, y_min, y_max,
                                              qc_chip_z, axis)
            else:
                # Only draw inductor
                self.render_junction_inductor(qgeom, x_min, x_max, y_min, y_max,
                                              qc_chip_z, axis)

    def render_junction_port(self, qgeom: pd.Series, xmin: float, xmax: float,
                             ymin: float, ymax: float, z: float, axis: str):
        """Render a junction as a port with a bounding box given by xmin/xmax
        and ymin/ymax, a height z, and a horizontal or vertical axis.

        Args:
            qgeom (pd.Series): GeoSeries of element properties.
            xmin (float): Smallest x coordinate
            xmax (float): Largest x coordinate
            ymin (float): Smallest y coordinate
            ymax (float): Largest y coordinate
            z (float): z coordinate
            axis (str): Orientation, either 'x' or 'y'
        """
        ansys_options = dict(transparency=0.0)
        qcomp = self.design._components[qgeom['component']].name
        qc_elt = get_clean_name(qgeom['name'])
        port_name = f'Port_{qcomp}_{qc_elt}'
        impedance = self.jj_lumped_ports[(qcomp, qc_elt)][0]
        # Draw rectangle for lumped port.
        self.logger.debug(f'Drawing a rectangle: {port_name}')
        poly_ansys = self.modeler.draw_rect_corner([xmin, ymin, z], xmax - xmin,
                                                   ymax - ymin, z,
                                                   **ansys_options)
        poly_ansys.make_lumped_port(axis,
                                    z0=str(impedance) + 'ohm',
                                    name=f'LumpPort_{qcomp}_{qc_elt}')
        self.modeler.rename_obj(poly_ansys, port_name)
        # Draw line for lumped port.
        if axis == 'x':
            ymid = (ymin + ymax) / 2
            start, end = [xmin, ymid, z], [xmax, ymid, z]
        elif axis == 'y':
            xmid = (xmin + xmax) / 2
            start, end = [xmid, ymin, z], [xmid, ymax, z]
        lump_line = self.modeler.draw_polyline([start, end],
                                               closed=False,
                                               **dict(color=(128, 0, 128)))
        lump_line = lump_line.rename(f'voltage_line_{port_name}')
        lump_line.show_direction = True

    def render_junction_inductor(self, qgeom: pd.Series, xmin: float,
                                 xmax: float, ymin: float, ymax: float,
                                 z: float, axis: str):
        """Render a junction as an inductor with a bounding box given by
        xmin/xmax and ymin/ymax, a height z, and a horizontal or vertical axis.

        Args:
            qgeom (pd.Series): GeoSeries of element properties.
            xmin (float): Smallest x coordinate
            xmax (float): Largest x coordinate
            ymin (float): Smallest y coordinate
            ymax (float): Largest y coordinate
            z (float): z coordinate
            axis (str): Orientation, either 'x' or 'y'
        """
        ansys_options = dict(transparency=0.0)
        qcomp = self.design._components[qgeom['component']].name
        qc_elt = get_clean_name(qgeom['name'])
        qc_name = 'Lj_' + qcomp
        inductor_name = f'{qc_name}{QAnsysRenderer.NAME_DELIM}{qc_elt}'
        # Draw rectangle for inductor.
        self.logger.debug(f'Drawing a rectangle: {inductor_name}')
        poly_ansys = self.modeler.draw_rect_corner([xmin, ymin, z], xmax - xmin,
                                                   ymax - ymin, 0,
                                                   **ansys_options)
        poly_ansys.make_rlc_boundary(axis,
                                     l=qgeom['hfss_inductance'],
                                     c=qgeom['hfss_capacitance'],
                                     r=qgeom['hfss_resistance'],
                                     name='Lj_' + inductor_name)
        self.modeler.rename_obj(poly_ansys, 'JJ_rect_' + inductor_name)
        self.assign_mesh.append('JJ_rect_' + inductor_name)
        # Draw line for inductor.
        if axis == 'x':
            ymid = (ymin + ymax) / 2
            start, end = [xmin, ymid, z], [xmax, ymid, z]
        elif axis == 'y':
            xmid = (xmin + xmax) / 2
            start, end = [xmid, ymin, z], [xmid, ymax, z]
        induc_line = self.modeler.draw_polyline([start, end],
                                                closed=False,
                                                **dict(color=(128, 0, 128)))
        induc_line = induc_line.rename('JJ_' + inductor_name + '_')
        induc_line.show_direction = True

    def metallize(self):
        """Assign metallic property to all shapes in self.assign_perfE list."""
        self.modeler.assign_perfect_E(self.assign_perfE)

    def add_drivenmodal_design(self, name: str, connect: bool = True):
        """
        (deprecated) use new_ansys_design()
        """
        self.logger.warning(
            'This method is deprecated. Change your scripts to use new_ansys_design()'
        )
        self.new_ansys_design(name, 'drivenmodal', connect)

    def activate_drivenmodal_design(self, name: str = "MetalHFSSDrivenModal"):
        """
        (deprecated) use activate_ansys_design()
        """
        self.logger.warning(
            'This method is deprecated. Change your scripts to use activate_ansys_design()'
        )
        self.activate_ansys_design(name, 'drivenmodal')

    def activate_drivenmodal_setup(self, setup_name_activate: str = None):
        """
        (deprecated) use activate_ansys_setup()
        """
        self.logger.warning(
            'This method is deprecated. Change your scripts to use activate_ansys_setup()'
        )
        self.activate_ansys_setup(setup_name_activate)

    def add_drivenmodal_setup(self,
                              name: str = None,
                              freq_ghz: int = None,
                              max_delta_s: float = None,
                              max_passes: int = None,
                              min_passes: int = None,
                              min_converged: int = None,
                              pct_refinement: int = None,
                              basis_order: int = None,
                              *args,
                              **kwargs):
        """Create a solution setup in Ansys HFSS Driven Modal. If user does
        not provide arguments, they will be obtained from default_setup dict.

        Args:
            name (str, optional): Name of driven modal setup. Defaults to None.
            freq_ghz (int, optional): Frequency in GHz. Defaults to None.
            max_delta_s (float, optional): Absolute value of maximum difference in scattering parameter S. Defaults to None.
            max_passes (int, optional): Maximum number of passes. Defaults to None.
            min_passes (int, optional): Minimum number of passes. Defaults to None.
            min_converged (int, optional): Minimum number of converged passes. Defaults to None.
            pct_refinement (int, optional): Percent refinement. Defaults to None.
            basis_order (int, optional): Basis order. Defaults to None.
        """
        dsu = self.default_setup.drivenmodal

        if not name:
            name = self.parse_value(dsu['name'])
        if not freq_ghz:
            freq_ghz = int(self.parse_value(dsu['freq_ghz']))
        if not max_delta_s:
            max_delta_s = float(self.parse_value(dsu['max_delta_s']))
        if not max_passes:
            max_passes = int(self.parse_value(dsu['max_passes']))
        if not min_passes:
            min_passes = int(self.parse_value(dsu['min_passes']))
        if not min_converged:
            min_converged = int(self.parse_value(dsu['min_converged']))
        if not pct_refinement:
            pct_refinement = int(self.parse_value(dsu['pct_refinement']))
        if not basis_order:
            basis_order = int(self.parse_value(dsu['basis_order']))

        if self.pinfo:
            if self.pinfo.design:
                return self.pinfo.design.create_dm_setup(
                    freq_ghz=freq_ghz,
                    name=name,
                    max_delta_s=max_delta_s,
                    max_passes=max_passes,
                    min_passes=min_passes,
                    min_converged=min_converged,
                    pct_refinement=pct_refinement,
                    basis_order=basis_order)

    def add_eigenmode_design(self, name: str, connect: bool = True):
        """
        (deprecated) use new_ansys_design()
        """
        self.logger.warning(
            'This method is deprecated. Change your scripts to use new_ansys_design()'
        )
        self.new_ansys_design(name, 'eigenmode', connect)

    def activate_eigenmode_design(self, name: str = "MetalHFSSEigenmode"):
        """
        (deprecated) use activate_ansys_design()
        """
        self.logger.warning(
            'This method is deprecated. Change your scripts to use activate_ansys_design()'
        )
        self.activate_ansys_design(name, 'eigenmode')

    def activate_eigenmode_setup(self, setup_name_activate: str = None):
        """
        (deprecated) use activate_ansys_setup()
        """
        self.logger.warning(
            'This method is deprecated. Change your scripts to use activate_ansys_setup()'
        )
        self.activate_ansys_setup(setup_name_activate)

    def add_eigenmode_setup(self,
                            name: str = None,
                            min_freq_ghz: int = None,
                            n_modes: int = None,
                            max_delta_f: float = None,
                            max_passes: int = None,
                            min_passes: int = None,
                            min_converged: int = None,
                            pct_refinement: int = None,
                            basis_order: int = None,
                            *args,
                            **kwargs):
        """Create a solution setup in Ansys HFSS Eigenmode.  If user does not
        provide arguments, they will be obtained from hfss_options dict.

        Args:
            name (str, optional): Name of eigenmode setup. Defaults to "Setup".
            min_freq_ghz (int, optional): Minimum frequency in GHz. Defaults to 1.
            n_modes (int, optional): Number of modes. Defaults to 1.
            max_delta_f (float, optional): Maximum difference in freq between consecutive passes. Defaults to 0.5.
            max_passes (int, optional): Maximum number of passes. Defaults to 10.
            min_passes (int, optional): Minimum number of passes. Defaults to 1.
            min_converged (int, optional): Minimum number of converged passes. Defaults to 1.
            pct_refinement (int, optional): Percent refinement. Defaults to 30.
            basis_order (int, optional): Basis order. Defaults to -1.
        """
        esu = self.default_setup.eigenmode

        if not name:
            name = self.parse_value(esu['name'])
        if not min_freq_ghz:
            min_freq_ghz = int(self.parse_value(esu['min_freq_ghz']))
        if not n_modes:
            n_modes = int(self.parse_value(esu['n_modes']))
        if not max_delta_f:
            max_delta_f = float(self.parse_value(esu['max_delta_f']))
        if not max_passes:
            max_passes = int(self.parse_value(esu['max_passes']))
        if not min_passes:
            min_passes = int(self.parse_value(esu['min_passes']))
        if not min_converged:
            min_converged = int(self.parse_value(esu['min_converged']))
        if not pct_refinement:
            pct_refinement = int(self.parse_value(esu['pct_refinement']))
        if not basis_order:
            basis_order = int(self.parse_value(esu['basis_order']))

        if self.pinfo:
            if self.pinfo.design:
                return self.pinfo.design.create_em_setup(
                    name=name,
                    min_freq_ghz=min_freq_ghz,
                    n_modes=n_modes,
                    max_delta_f=max_delta_f,
                    max_passes=max_passes,
                    min_passes=min_passes,
                    min_converged=min_converged,
                    pct_refinement=pct_refinement,
                    basis_order=basis_order)

    def edit_eigenmode_setup(self, setup_args: Dict):
        """User can pass key/values to edit the setup for active eigenmode setup.

        Args:
            setup_args (Dict): a Dict with possible keys/values.

        **setup_args** dict contents:
            * name (str, optional): Name of eigenmode setup. Defaults to "Setup".
            * min_freq_ghz (int, optional): Minimum frequency in GHz. Defaults to 1.
            * n_modes (int, optional): Number of modes. Defaults to 1.
            * max_delta_f (float, optional): Maximum difference in freq between consecutive passes. Defaults to 0.5.
            * max_passes (int, optional): Maximum number of passes. Defaults to 10.
            * pct_refinement (int, optional): Percent refinement. Defaults to 30.
            * basis_order (int, optional): Basis order. Defaults to -1.

            Note, that these two are currently NOT implemented:
            Ansys API named EditSetup not documented for HFSS, and
            self.pinfo.setup does not have all the property variables used for Setup.
            * min_passes (int, optional): Minimum number of passes. Defaults to 1.
            * min_converged (int, optional): Minimum number of converged passes. Defaults to 1.
        """

        if self.pinfo:
            if self.pinfo.project:
                if self.pinfo.design:
                    if self.pinfo.design.solution_type == 'Eigenmode':
                        if self.pinfo.setup_name != setup_args.name:
                            self.design.logger.warning(
                                f'The name of active setup={self.pinfo.setup_name} does not match'
                                f'the name of of setup_args.name={setup_args.name}. '
                                f'To use this method, activate the desired Setup before editing it. The '
                                f'setup_args was not used to update the active Setup.'
                            )
                            return

                        for key, value in setup_args.items():
                            if key == "name":
                                continue  #Checked for above.
                            if key == "n_modes":
                                #EditSetup  not documented, this is just attempt to use.
                                #args_editsetup = ["NAME:" + setup_args.name,["NumModes:=", setup_args.n_modes]]
                                #self.pinfo.setup._setup_module.EditSetup([setup_args.name, args_editsetup])
                                if value < 0 or value > 20 or not isinstance(
                                        value, int):
                                    self.logger.warning(
                                        f'Value of n_modes={value} must be integer from 1 to 20.'
                                    )
                                else:
                                    self.pinfo.setup.n_modes = value
                                    continue
                            if key == "min_freq_ghz":
                                if not isinstance(value, int):
                                    self.logger.warning(
                                        'The value for min_freq_ghz should be an int. '
                                        f'The present value is {value}.')
                                else:
                                    self.pinfo.setup.min_freq = f'{value}GHz'
                                    continue
                            if key == 'max_delta_f':
                                if not isinstance(value, float):
                                    self.logger.warning(
                                        'The value for max_delta_f should be float. '
                                        f'The present value is {value}.')
                                else:
                                    self.pinfo.setup.delta_f = value
                                    continue
                            if key == 'max_passes':
                                if not isinstance(value, int):
                                    self.logger.warning(
                                        'The value for max_passes should be an int. '
                                        f'The present value is {value}.')
                                else:
                                    self.pinfo.setup.passes = value
                                    continue
                            if key == 'pct_refinement':
                                if not isinstance(value, int):
                                    self.logger.warning(
                                        'The value for pct_refinement should be an int. '
                                        f'The present value is {value}.')
                                else:
                                    self.pinfo.setup.pct_refinement = value
                                    continue
                            if key == 'basis_order':
                                if not isinstance(value, int):
                                    self.logger.warning(
                                        'The value for basis_order should be an int. '
                                        f'The present value is {value}.')
                                else:
                                    self.pinfo.setup.basis_order = value
                                    continue

                            self.design.logger.warning(
                                f'In setup_args, key={key}, value={value} is not in pinfo.setup, '
                                'the key/value pair from setup_args not added to Setup in Ansys.'
                            )

                    else:
                        self.logger.warning(
                            'The design does not have solution type as "Eigenmode". The Setup not updated.'
                        )
                else:
                    self.logger.warning(
                        'A design is not in active project. The Setup not updated.'
                    )
            else:
                self.logger.warning(
                    "Project not available, have you opened a project? Setup not updated."
                )
        else:
            self.logger.warning(
                "Have you run connect_ansys()?  "
                "Cannot find a reference to Ansys in QRenderer. Setup not updated. "
            )

    def edit_drivenmodal_setup(self, setup_args: Dict):
        """User can pass key/values to edit the setup for active driven modal setup.

        Args:
            setup_args (Dict): a Dict with possible keys/values.

        **setup_args** dict contents:
            * name (str, optional): Name of eigenmode setup. Defaults to "Setup".
            * freq_ghz (int, optional): Minimum frequency in GHz. Defaults to 1.
            * max_passes (int, optional): Maximum number of passes. Defaults to 10.
            * pct_refinement (int, optional): Percent refinement. Defaults to 30.
            * basis_order (int, optional): Basis order. Defaults to -1 (1 is "Mixed Order").

            Note, that these three are currently NOT implemented:
            Ansys API named EditSetup not documented for HFSS, and
            self.pinfo.setup does not have all the property variables used for Setup.
            * max_delta_s (float, optional): Absolute value of maximum difference in scattering parameter S. Defaults to 0.1.
            * min_passes (int, optional): Minimum number of passes. Defaults to 1.
            * min_converged (int, optional): Minimum number of converged passes. Defaults to 1.
        """

        if self.pinfo:
            if self.pinfo.project:
                if self.pinfo.design:
                    if self.pinfo.design.solution_type == 'DrivenModal':
                        if self.pinfo.setup_name != setup_args.name:
                            self.design.logger.warning(
                                f'The name of active setup={self.pinfo.setup_name} does not match'
                                f'the name of of setup_args.name={setup_args.name}. '
                                f'To use this method, activate the desired Setup before editing it. The '
                                f'setup_args was not used to update the active Setup.'
                            )
                            return

                        for key, value in setup_args.items():
                            if key == "name":
                                continue  #Checked for above.
                            if key == "freq_ghz":
                                if not isinstance(value, float):
                                    self.logger.warning(
                                        'The value for freq_ghz should be an float. '
                                        f'The present value is {value}.')
                                else:
                                    self.pinfo.setup.solution_freq = f'{value}GHz'
                                    continue
                            if key == 'max_passes':
                                if not isinstance(value, int):
                                    self.logger.warning(
                                        'The value for passes should be an int. '
                                        f'The present value is {value}.')
                                else:
                                    self.pinfo.setup.passes = value
                                    continue
                            if key == 'pct_refinement':
                                if not isinstance(value, int):
                                    self.logger.warning(
                                        'The value for pct_refinement should be an int. '
                                        f'The present value is {value}.')
                                else:
                                    self.pinfo.setup.pct_refinement = value
                                    continue
                            if key == 'basis_order':
                                if not isinstance(value, int):
                                    self.logger.warning(
                                        'The value for basis_order should be an int. '
                                        f'The present value is {value}.')
                                else:
                                    self.pinfo.setup.basis_order = value
                                    continue

                            self.design.logger.warning(
                                f'In setup_args, key={key}, value={value} is not in pinfo.setup, '
                                'the key/value pair from setup_args not added to Setup in Ansys.'
                            )

                    else:
                        self.logger.warning(
                            'The design does not have solution type as "Driven Modal". The Setup not updated.'
                        )
                else:
                    self.logger.warning(
                        'A design is not in active project. The Setup not updated.'
                    )
            else:
                self.logger.warning(
                    "Project not available, have you opened a project? Setup not updated."
                )
        else:
            self.logger.warning(
                "Have you run connect_ansys()?  "
                "Cannot find a reference to Ansys in QRenderer. Setup not updated. "
            )

    def set_mode(self, mode: int, setup_name: str):
        """Set the eigenmode in pyEPR for a design with solution_type set to
        Eigenmode.

        Args:
            mode (int): Identify a mode from 1 to n_modes.
            setup_name (str): Select a setup from the active design.
        """
        if self.pinfo:
            if self.pinfo.project:
                if self.pinfo.design:
                    # double parent, becasue self.pinfo.design does not work
                    o_desktop = self.pinfo.design.parent.parent._desktop
                    o_project = o_desktop.SetActiveProject(
                        self.pinfo.project_name)
                    o_design = o_project.GetActiveDesign()
                    if o_design.GetSolutionType() == 'Eigenmode':
                        # The set_mode() method is in HfssEMDesignSolutions
                        #  class in pyEPR.
                        # The class HfssEMDesignSolutions is instantiated by
                        #  get_setup() and create_em_setup().
                        setup = self.pinfo.get_setup(setup_name)
                        if 0 < int(mode) <= int(setup.n_modes):
                            setup_solutions = setup.get_solutions()
                            if setup_solutions:
                                setup_solutions.set_mode(mode)
                            else:
                                self.logger.warning(
                                    'Not able to get setup_solutions, '
                                    'the mode was not set.')
                        else:
                            self.logger.warning(
                                f'The requested mode={mode} is not a valid '
                                f'(1 to {setup.n_modes}) selection. '
                                'The mode was not set.')
                    else:
                        self.logger.warning(
                            'The design does not have solution type as '
                            '"Eigenmode". The mode was not set.')
                else:
                    self.logger.warning('A design is not in active project. '
                                        'The mode was not set.')
            else:
                self.logger.warning(
                    "Project not available, have you opened a project? "
                    "The mode was not set.")
        else:
            self.logger.warning(
                "Have you run connect_ansys()?  "
                "Cannot find a reference to Ansys in QRenderer.  "
                "The mode was not set.")

    def analyze_setup(self, setup_name: str):
        """Run a specific solution setup in Ansys HFSS.

        Args:
            setup_name (str): Name of setup.
        """
        if self.pinfo:
            setup = self.pinfo.get_setup(setup_name)
            setup.analyze(setup_name)

    def add_sweep(self,
                  setup_name="Setup",
                  start_ghz=2.0,
                  stop_ghz=8.0,
                  count=101,
                  step_ghz=None,
                  name="Sweep",
                  type="Fast",
                  save_fields=False):
        """Add a frequency sweep to a driven modal setup.

        Args:
            setup_name (str, optional): Name of driven modal simulation setup.
                                    Defaults to "Setup".
            start_ghz (float, optional): Starting frequency of sweep in GHz.
                                    Defaults to 2.0.
            stop_ghz (float, optional): Ending frequency of sweep in GHz.
                                    Defaults to 8.0.
            count (int, optional): Total number of frequencies.
                                    Defaults to 101.
            step_ghz (float, optional): Difference between adjacent
                                    frequencies. Defaults to None.
            name (str, optional): Name of sweep. Defaults to "Sweep".
            type (str, optional): Type of sweep. Defaults to "Fast".
            save_fields (bool, optional): Whether or not to save fields.
                                Defaults to False.
        """
        if self.pinfo:
            setup = self.pinfo.get_setup(setup_name)
            return setup.insert_sweep(start_ghz=start_ghz,
                                      stop_ghz=stop_ghz,
                                      count=count,
                                      step_ghz=step_ghz,
                                      name=name,
                                      type=type,
                                      save_fields=save_fields)

    def analyze_sweep(self, sweep_name: str, setup_name: str):
        """Analyze a single sweep within the setup.

        Args:
            sweep_name (str): Name of sweep to analyze.
            setup_name (str): Name of setup to analyze.
        """
        if self.pinfo:
            setup = self.pinfo.get_setup(setup_name)
            sweep = setup.get_sweep(sweep_name)
            sweep.analyze_sweep()
            self.current_sweep = sweep

    def get_params(self, param_name: Union[list, None] = None):
        """Get one or more parameters (S, Y, or Z) as a function of frequency.

        Args:
            param_name (Union[list, None], optional): Parameters to obtain. Defaults to None.
        """
        if self.current_sweep:

            freqs, Pcurves = self.current_sweep.get_network_data(param_name)
            Pparams = pd.DataFrame(Pcurves,
                                   columns=freqs / 1e9,
                                   index=param_name).transpose()
        return freqs, Pcurves, Pparams

    # yapf: disable
    def get_all_Pparms_matrices(self, matrix_size: int) -> Tuple[
            Union[pd.core.frame.DataFrame, None],
            Union[pd.core.frame.DataFrame, None],
            Union[pd.core.frame.DataFrame, None]]:
        #yapf: enable
        '''
        S = scattering matrix, Y = Admittance, Z= impedance.

        matrix_size should be 1 or larger.
        This method will get the entire Scattering matrix based on matrix_size.

        Example:'S21'
        S matrix: SAB means, excite B, measure A
        '''
        s_param_name = []
        y_param_name = []
        z_param_name = []
        if matrix_size < 1:
            return None, None, None
        for excite in range(1, matrix_size + 1):
            for measure in range(1, matrix_size + 1):
                s_param_name.append(f'S{measure}{excite}')
                y_param_name.append(f'Y{measure}{excite}')
                z_param_name.append(f'Z{measure}{excite}')
        dummy_freqs, dummy_Pcurves, S_Pparams = self.get_params(s_param_name)
        dummy_freqs, dummy_Pcurves, Y_Pparams = self.get_params(y_param_name)
        dummy_freqs, dummy_Pcurves, Z_Pparams = self.get_params(z_param_name)

        return S_Pparams, Y_Pparams, Z_Pparams

    def plot_params(self, param_name: Union[list, None] = None):
        """Plot one or more parameters (S, Y, or Z) as a function of frequency.
        S = scattering matrix, Y = Admittance, Z= impedance.

        Args:
            param_name (Union[list, None], optional): Parameters to plot. Defaults to None.
        """
        freqs, Pcurves, Pparams = self.get_params(param_name)
        if Pparams is not None:
            fig, axs = plt.subplots(1, 2, figsize=(10, 6))
            Pparams.apply(lambda x: 20 * np.log10(np.abs(x))).plot(ax=axs[0])
            axs[0].set_xlabel('Frequency (GHz)')
            axs[0].set_ylabel('Amplitude (dB)')
            Pparams.apply(lambda x: np.angle(x)).plot(ax=axs[1])
            axs[1].set_xlabel('Frequency (GHz)')
            axs[1].set_ylabel('Phase (rad.)')
            for ax in axs:
                ax.autoscale()
        return Pparams, fig

    def distributed_analysis(self):
        """Returns class containing info on Hamiltonian parameters from HFSS
        simulation.

        Returns:
            DistributedAnalysis: A  class from pyEPR which does DISTRIBUTED ANALYSIS of layout
            and microwave results.  It is the main computation class & interface with HFSS.
            This class defines a DistributedAnalysis object which calculates
            and saves Hamiltonian parameters from an HFSS simulation.
            It allows one to calculate dissipation.
        """
        if self.pinfo:
            return epr.DistributedAnalysis(self.pinfo)

    def get_convergences(self, variation: str = None):
        """Get convergence for convergence_t, convergence_f, and text from GUI for solution data.

        Args:
            variation (str, optional):  Information from pyEPR; variation should be in the form
            variation = "scale_factor='1.2001'". Defaults to None.

        Returns:
            tuple[pandas.DataFrame, pandas.DataFrame, str]:
            1st DataFrame: Convergence_t
            2nd DataFrame: Convergence_f
            3rd str: Text from GUI of solution data.
        """
        if self.pinfo:
            convergence_t, text = self.pinfo.setup.get_convergence(variation)
            convergence_f = self.get_f_convergence([])  # TODO; Fix variation []
            return convergence_t, convergence_f, text

    def plot_convergences(self,
                          variation: str = None,
                          fig: mpl.figure.Figure = None):
        """
        (deprecated) use EPRanalysis.plot_convergences()
        """
        print(
            'This method is deprecated. Change your scripts to use the equivalent method '
            'plot_convergence() that has been moved inside the EPRanalysis class.'
        )

    def get_f_convergence(self, variation: str = None, save_csv: bool = True):
        """Create a report inside HFSS to plot the converge of frequency and
        style it. Saves report to csv file.

        .. code-block:: text

                re(Mode(1)) [g]	re(Mode(2)) [g]	re(Mode(3)) [g]
            Pass []
            1	4.643101	4.944204	5.586289
            2	5.114490	5.505828	6.242423
            3	5.278594	5.604426	6.296777

        Args:
            variation ('str', optional): Information from pyEPR; variation should be in the form
                variation = "scale_factor='1.2001'". Defaults to None.
            save_csv (bool, optional): Save to file? Defaults to True.

        Returns:
            pd.DataFrame: Returns a convergence vs pass number of the eigenemode frequencies.
        """
        if self.pinfo:
            o_design = self.pinfo.design
            setup = self.pinfo.setup

            if not o_design.solution_type == 'Eigenmode':
                return None

            report = o_design._reporter  # reporter OModule for Ansys

            # Create report
            n_modes = int(setup.n_modes)
            ycomp = [f"re(Mode({i}))" for i in range(1, 1 + n_modes)]

            params = ["Pass:=", ["All"]] + variation
            report_name = "Freq. vs. pass"
            if report_name in report.GetAllReportNames():
                report.DeleteReports([report_name])

            solutions = setup.get_solutions()
            solutions.create_report(report_name,
                                    "Pass",
                                    ycomp,
                                    params,
                                    pass_name='AdaptivePass')

            # Properties of lines
            curves = [
                f"{report_name}:re(Mode({i})):Curve1"
                for i in range(1, 1 + n_modes)
            ]
            set_property(report, 'Attributes', curves, 'Line Width', 3)
            set_property(report, 'Scaling', f"{report_name}:AxisY1",
                         'Auto Units', False)
            set_property(report, 'Scaling', f"{report_name}:AxisY1", 'Units',
                         'g')
            set_property(report, 'Legend', f"{report_name}:Legend",
                         'Show Solution Name', False)

            if save_csv:  # Save
                try:
                    path = Path().absolute(
                    ) / 'hfss_eig_f_convergence.csv'  # TODO: Determine better path
                    report.ExportToFile(report_name, path)
                    self.logger.info(f'Saved convergences to {path}')
                    return pd.read_csv(path, index_col=0)
                except Exception as e:
                    self.logger.error(
                        f"Error could not save and export hfss plot to {path}.\
                                Is the plot made in HFSS with the correct name.\
                                Check the HFSS error window. \t Error =  {e}")

            return None



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Tunable Coupler, as shown in Phys. Rev. Research 2, 033447 (2020).
"""

from math import sin, cos
import numpy as np
from qiskit_metal import draw, Dict
from qiskit_metal.qlibrary.core import BaseQubit


class TunableCoupler02(BaseQubit):
    """One of the tunable couplers

    .. TODO: add apidocs/TunableCoupler02.png once available

    .. meta::
        :description: Tunable Coupler 2

    BaseQubit Default Options:
        * connection_pads: empty Dict -- Currently not used, connection count is static. (WIP)
    Default Options:
        * pocket_width: '3mm' -- The width (x-axis) of the transmon pocket
        * pocket_height: '2m' -- The height (y-axis) of the transmon pocket
        * pos_x: '0mm' -- the x-coordinate location of the JJ in the middle of the qubit
        * pos_y: '0mm' -- the y-coordinate location of the JJ in the middle of the qubit
        * orientation: '0' -- the angle of rotation
        * pad_radius: '0.3mm' -- the radius of the two circular pads on either side of the JJ
        * bus_width: '0.1mm' -- the width of the CPW resonator in the coupler
        * bus_inner_length: '0.5mm' -- length of the CPW between the JJ and the circular pad
        * bus_outer_length: '0.5mm' -- length of the CPW between the circular pad and the in/out pin
        * bus_JJ_height: '0.5mm' -- vertical distance between the main cpw line and the offset JJ
        * JJ_width: '0.1mm' -- the width of the JJ
        * JJ_length: '0.4mm' -- the length of the JJ
        * fbl_width: '0.1mm' -- the width of the flux bias line drawn near the JJ
        * fbl_length: '0.5mm' -- the length of the flux bias line drawn near the JJ
        * fbl_offset: '0.1mm' -- the lateral offset between the flux bias line and the JJ
        * layer: '1' -- the default design layer
    """

    default_options = Dict(pocket_width='3mm',
                           pocket_height='2mm',
                           pos_x='0mm',
                           pos_y='0mm',
                           orientation='0',
                           pad_radius='0.3mm',
                           bus_width='0.1mm',
                           bus_inner_length='0.5mm',
                           bus_outer_length='0.5mm',
                           bus_JJ_height='0.5mm',
                           JJ_width='0.1mm',
                           JJ_length='0.4mm',
                           fbl_width='0.1mm',
                           fbl_length='0.5mm',
                           fbl_offset='0.1mm',
                           layer='1')

    component_metadata = Dict(short_name='Pocket',
                              _qgeometry_table_path='True',
                              _qgeometry_table_poly='True',
                              _qgeometry_table_junction='True')

    TOOLTIP = """One of the tunable couplers"""

    def make(self):
        """Builds the component."""
        """Convert self.options into QGeometry."""
        p = self.parse_options()  # Parse the string options into numbers
        # draw the Josephson Junction
        JJ = draw.LineString([
            (-0.5 * p.JJ_length, p.bus_JJ_height - 0.5 * p.JJ_width),
            (0.5 * p.JJ_length, p.bus_JJ_height - 0.5 * p.JJ_width)
        ])
        # draw the individual components of the CPW resonator
        bus_vertical_left = draw.rectangle(
            p.bus_width, p.bus_JJ_height,
            -0.5 * p.JJ_length - 0.5 * p.bus_width, 0.5 * p.bus_JJ_height)
        bus_vertical_right = draw.rectangle(
            p.bus_width, p.bus_JJ_height, 0.5 * p.JJ_length + 0.5 * p.bus_width,
            0.5 * p.bus_JJ_height)
        bus_left = draw.rectangle(
            p.bus_inner_length + p.bus_outer_length + 2.0 * p.pad_radius,
            p.bus_width, -0.5 * p.JJ_length - p.bus_width - 0.5 *
            (p.bus_inner_length + p.bus_outer_length + 2.0 * p.pad_radius),
            0.5 * p.bus_width)
        bus_right = draw.rectangle(
            p.bus_inner_length + p.bus_outer_length + 2.0 * p.pad_radius,
            p.bus_width, 0.5 * p.JJ_length + p.bus_width + 0.5 *
            (p.bus_inner_length + p.bus_outer_length + 2.0 * p.pad_radius),
            0.5 * p.bus_width)
        # draw the circular charge pads
        left_pad = draw.Point(
            -0.5 * p.JJ_length - p.bus_width - p.bus_inner_length -
            0.5 * p.pad_radius, 0.5 * p.bus_width).buffer(p.pad_radius)
        right_pad = draw.Point(
            0.5 * p.JJ_length + p.bus_width + p.bus_inner_length +
            0.5 * p.pad_radius, 0.5 * p.bus_width).buffer(p.pad_radius)
        # draw the flux bias line
        fbl = draw.rectangle(
            p.fbl_width, p.fbl_length,
            -0.5 * p.JJ_length - p.bus_width - p.fbl_offset - 0.5 * p.fbl_width,
            p.bus_JJ_height + 0.5 * p.fbl_length)
        # draw the pocket surrounding the qubit
        pocket = draw.rectangle(0.0, 0.0, p.pocket_width, p.pocket_height)
        # Translate and rotate all shapes
        objects = [
            JJ, bus_vertical_left, bus_vertical_right, bus_left, bus_right,
            left_pad, right_pad, fbl, pocket
        ]
        objects = draw.rotate(objects, p.orientation, origin=(0, 0))
        objects = draw.translate(objects, xoff=p.pos_x, yoff=p.pos_y)
        [
            JJ, bus_vertical_left, bus_vertical_right, bus_left, bus_right,
            left_pad, right_pad, fbl, pocket
        ] = objects
        # give each poly a name to send to qgeometry
        geom_jj = {'poly1': JJ}
        geom_bus_vertical_left = {'poly2': bus_vertical_left}
        geom_bus_vertical_right = {'poly3': bus_vertical_right}
        geom_bus_left = {'poly4': bus_left}
        geom_bus_right = {'poly5': bus_right}
        geom_left_pad = {'poly6': left_pad}
        geom_right_pad = {'poly7': right_pad}
        geom_fbl = {'poly8': fbl}
        #geom_pocket = {'poly9': pocket}
        # add to qgeometry
        self.add_qgeometry('junction',
                           geom_jj,
                           layer=p.layer,
                           subtract=False,
                           width=p.JJ_width)
        self.add_qgeometry('poly',
                           geom_bus_vertical_left,
                           layer=p.layer,
                           subtract=False)
        self.add_qgeometry('poly',
                           geom_bus_vertical_right,
                           layer=p.layer,
                           subtract=False)
        self.add_qgeometry('poly', geom_bus_left, layer=p.layer, subtract=False)
        self.add_qgeometry('poly',
                           geom_bus_right,
                           layer=p.layer,
                           subtract=False)
        self.add_qgeometry('poly', geom_left_pad, layer=p.layer, subtract=False)
        self.add_qgeometry('poly',
                           geom_right_pad,
                           layer=p.layer,
                           subtract=False)
        self.add_qgeometry('poly', geom_fbl, layer=p.layer, subtract=False)

        #self.add_qgeometry('poly', geom_pocket, layer=1, subtract=True)
        ###########################################################################
        # Add Qpin connections
        # define a function that both rotates and translates the qpin coordinates
        def qpin_rotate_translate(x):
            y = list(x)
            z = [0.0, 0.0]
            z[0] = y[0] * cos(p.orientation * 3.14159 / 180) - y[1] * sin(
                p.orientation * 3.14159 / 180)
            z[1] = y[0] * sin(p.orientation * 3.14159 / 180) + y[1] * cos(
                p.orientation * 3.14159 / 180)
            z[0] = z[0] + p.pos_x
            z[1] = z[1] + p.pos_y
            x = (z[0], z[1])
            return x

        # Right-hand pin
        qp1b = (0.5 * p.JJ_length + p.bus_width + p.bus_inner_length +
                p.bus_outer_length + 2.0 * p.pad_radius, 0.5 * p.bus_width)
        qp1a = (0.5 * p.JJ_length + p.bus_width + 0.9 * p.bus_inner_length +
                p.bus_outer_length + 2.0 * p.pad_radius, 0.5 * p.bus_width)
        qp1a = qpin_rotate_translate(qp1a)
        qp1b = qpin_rotate_translate(qp1b)
        self.add_pin('pin1',
                     points=np.array([qp1a, qp1b]),
                     width=0.01,
                     input_as_norm=True)
        # Left-hand pin
        qp2b = (-0.5 * p.JJ_length - p.bus_width - p.bus_inner_length -
                p.bus_outer_length - 2.0 * p.pad_radius, 0.5 * p.bus_width)
        qp2a = (-0.5 * p.JJ_length - p.bus_width - 0.9 * p.bus_inner_length -
                p.bus_outer_length - 2.0 * p.pad_radius, 0.5 * p.bus_width)
        qp2a = qpin_rotate_translate(qp2a)
        qp2b = qpin_rotate_translate(qp2b)
        self.add_pin('pin2',
                     points=np.array([qp2a, qp2b]),
                     width=0.01,
                     input_as_norm=True)
        # Flux Bias Line pin
        qp3b = (-0.5 * p.JJ_length - p.bus_width - p.fbl_offset -
                0.5 * p.fbl_width, p.bus_JJ_height + p.fbl_length)
        qp3a = (-0.5 * p.JJ_length - p.bus_width - p.fbl_offset -
                0.5 * p.fbl_width, p.bus_JJ_height + 0.9 * p.fbl_length)
        qp3a = qpin_rotate_translate(qp3a)
        qp3b = qpin_rotate_translate(qp3b)
        self.add_pin('fbl',
                     points=np.array([qp3a, qp3b]),
                     width=0.01,
                     input_as_norm=True)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import pandas as pd
from pint import UnitRegistry

from pyEPR.calcs.convert import Convert

from qiskit_metal.designs import QDesign  # pylint: disable=unused-import
from qiskit_metal.analyses.core import QAnalysis
from qiskit_metal.analyses.simulation import LumpedElementsSim
from qiskit_metal import Dict, config

if not config.is_building_docs():
    from .lumped_capacitive import extract_transmon_coupled_Noscillator


# TODO: eliminate every reference to "renderer" in this file
class LOManalysis(QAnalysis):
    """Performs Lumped Oscillator Model analysis on a simulated or user-provided capacitance matrix.

    Default Setup:
        * junctions (Dict)
            * Lj (float): Junction inductance (in nH)
            * Cj (float): Junction capacitance (in fF)
        * freq_readout (float): Coupling readout frequency (in GHz).
        * freq_bus (Union[list, float]): Coupling bus frequencies (in GHz).
            * freq_bus can be a list with the order they appear in the capMatrix.

    Data Labels:
        * lumped_oscillator (pd.DataFrame): Lumped oscillator result at the last simulation pass
        * lumped_oscillator_all (dict): of pd.DataFrame. Lumped oscillator resulting
            at every pass of the simulation

    """
    default_setup = Dict(junctions=Dict(Lj=12, Cj=2),
                         freq_readout=7.0,
                         freq_bus=[6.0, 6.2])
    """Default setup."""

    # supported labels for data generated from the simulation
    data_labels = ['lumped_oscillator', 'lumped_oscillator_all']
    """Default data labels."""

    def __init__(self, design: 'QDesign' = None, renderer_name: str = None):
        """Initialize the Lumped Oscillator Model analysis.

        Args:
            design (QDesign): Pointer to the main qiskit-metal design.
                Used to access the QRenderer. Defaults to None.
            renderer_name (str, optional): Which renderer to use. Valid entries: 'q3d'.
                Defaults to None.
        """
        # QAnalysis are expected to either run simulation or use pre-saved sim outputs
        # we use a Dict() to store the sim outputs previously saved. Its key names need
        # to match those found in the correspondent simulation class.
        self.sim = Dict() if renderer_name is None else LumpedElementsSim(
            design, renderer_name)
        super().__init__()

    @property
    def lumped_oscillator(self) -> dict:
        """Getter

        Returns:
            dict: Lumped oscillator result at the last simulation pass.
        """
        return self.get_data('lumped_oscillator')

    @lumped_oscillator.setter
    def lumped_oscillator(self, data: dict):
        """Setter

        Args:
            data (dict): Lumped oscillator result at the last simulation pass.
        """
        if not isinstance(data, dict):
            self.logger.warning(
                'Unsupported type %s. Only accepts dict. Please try again.',
                {type(data)})
            return
        self.set_data('lumped_oscillator', data)

    @property
    def lumped_oscillator_all(self) -> pd.DataFrame:
        """Getter

        Returns:
            pd.DataFrame: each line corresponds to a simulation pass number
                and the remainder of the data is the respective lump oscillator information.
        """
        return self.get_data('lumped_oscillator_all')

    @lumped_oscillator_all.setter
    def lumped_oscillator_all(self, data: pd.DataFrame):
        """Setter

        Args:
            data (pd.DataFrame): each line corresponds to a simulation pass number
                and the remainder of the data is the respective lump oscillator information.
        """
        if not isinstance(data, pd.DataFrame):
            self.logger.warning(
                'Unsupported type %s. Only accepts pd.DataFrame. Please try again.',
                {type(data)})
            return
        self.set_data('lumped_oscillator_all', data)

    def run(self, *args, **kwargs):
        """Executes sequentially the system capacitance simulation (if a renderer was provided
        at creation of this object) and lom extraction by executing the methods
        LumpedElementsSim.run_sim(`*args`, `**kwargs`) and LOManalysis.run_lom().
        For input parameter, see documentation for LumpedElementsSim.run_sim().

        Returns:
            (dict): Pass numbers (keys) and respective lump oscillator information (values).
        """
        if isinstance(self.sim, LumpedElementsSim):
            self.sim.run(*args, **kwargs)
        return self.run_lom()

    def run_lom(self):
        """Executes the lumped oscillator extraction from the capacitance matrix,
        and based on the setup values.

        Returns:
            dict: Pass numbers (keys) and their respective capacitance matrices (values).
        """
        # wipe data from the previous run (if any)
        self.clear_data()

        s = self.setup

        if not isinstance(self.sim.capacitance_matrix, pd.DataFrame):
            if self.sim.capacitance_matrix == {}:
                self.logger.warning(
                    'Please initialize the capacitance_matrix before executing this method.'
                    '`self.sim.capacitance_matrix = pd.DataFrame(...)`')
                return
            if self.sim.capacitance_all_passes == {}:
                self.sim.capacitance_all_passes[
                    1] = self.sim.capacitance_matrix.values

        ureg = UnitRegistry()
        ic_amps = Convert.Ic_from_Lj(s.junctions.Lj, 'nH', 'A')
        cj = ureg(f'{s.junctions.Cj} fF').to('farad').magnitude
        fread = ureg(f'{s.freq_readout} GHz').to('GHz').magnitude
        fbus = [ureg(f'{freq} GHz').to('GHz').magnitude for freq in s.freq_bus]

        # derive number of coupling pads
        num_cpads = 2
        if isinstance(fread, list):
            num_cpads += len(fread) - 1
        if isinstance(fbus, list):
            num_cpads += len(fbus) - 1

        # get the LOM for every pass
        all_res = {}
        for idx_cmat, df_cmat in self.sim.capacitance_all_passes.items():
            res = extract_transmon_coupled_Noscillator(
                df_cmat,
                ic_amps,
                cj,
                num_cpads,
                fbus,
                fread,
                g_scale=1,
                print_info=bool(
                    idx_cmat == len(self.sim.capacitance_all_passes)))
            all_res[idx_cmat] = res
        self.lumped_oscillator = all_res[len(self.sim.capacitance_all_passes)]
        all_res = pd.DataFrame(all_res).transpose()
        all_res['χr MHz'] = abs(all_res['chi_in_MHz'].apply(lambda x: x[0]))
        all_res['gr MHz'] = abs(all_res['gbus'].apply(lambda x: x[0]))
        self.lumped_oscillator_all = all_res
        return self.lumped_oscillator_all

    def plot_convergence(self, *args, **kwargs):
        """Plots alpha and frequency versus pass number, as well as convergence of delta (in %).

        It accepts the same inputs as run_lom(), to allow regenerating the LOM
        results before plotting them.
        """
        if self.lumped_oscillator_all is None or args or kwargs:
            self.run_lom(*args, **kwargs)
        # TODO: copy plot_convergence_main() from pyEPR and move it here
        self.sim.renderer.plot_convergence_main(self.lumped_oscillator_all)

    def plot_convergence_chi(self, *args, **kwargs):
        """Plot convergence of chi and g, both in MHz, as a function of pass number.

        It accepts the same inputs as run_lom(), to allow regenerating the LOM
        results before plotting them.
        """
        if self.lumped_oscillator_all is None or args or kwargs:
            self.run_lom(*args, **kwargs)
        # TODO: copy plot_convergence_main() from pyEPR and move it here
        self.sim.renderer.plot_convergence_chi(self.lumped_oscillator_all)

    def load_simulation_data(self, data_name: str, data):
        """Load simulation data for the following analysis. This will override any data found

        Args:
            data_name (str): name of the variable
            data (Any): simulation output
        """
        self.sim[data_name] = data



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './endcap_q3d_ui.ui',
# licensing of './endcap_q3d_ui.ui' applies.
#
# Created: Sat Jun 19 22:02:29 2021
#      by: pyside6-uic  running on PySide2 5.13.2
#
# WARNING! All changes made in this file will be lost!

from PySide6 import QtCore, QtGui, QtWidgets


class Ui_mainWindow(object):

    def setupUi(self, mainWindow):
        mainWindow.setObjectName("mainWindow")
        mainWindow.resize(322, 530)
        self.centralwidget = QtWidgets.QWidget(mainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.label = QtWidgets.QLabel(self.centralwidget)
        self.label.setGeometry(QtCore.QRect(10, 10, 271, 20))
        self.label.setObjectName("label")
        self.verticalLayoutWidget = QtWidgets.QWidget(self.centralwidget)
        self.verticalLayoutWidget.setGeometry(QtCore.QRect(100, 450, 131, 32))
        self.verticalLayoutWidget.setObjectName("verticalLayoutWidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.verticalLayoutWidget)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.renderButton = QtWidgets.QPushButton(self.verticalLayoutWidget)
        self.renderButton.setObjectName("renderButton")
        self.verticalLayout.addWidget(self.renderButton)
        self.tableWidget = QtWidgets.QTableWidget(self.centralwidget)
        self.tableWidget.setGeometry(QtCore.QRect(10, 30, 301, 421))
        self.tableWidget.setSizeAdjustPolicy(
            QtWidgets.QAbstractScrollArea.AdjustIgnored)
        self.tableWidget.setObjectName("tableWidget")
        self.tableWidget.setColumnCount(0)
        self.tableWidget.setRowCount(0)
        mainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar()
        self.menubar.setGeometry(QtCore.QRect(0, 0, 322, 22))
        self.menubar.setObjectName("menubar")
        mainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(mainWindow)
        self.statusbar.setObjectName("statusbar")
        mainWindow.setStatusBar(self.statusbar)

        self.retranslateUi(mainWindow)
        QtCore.QObject.connect(self.renderButton, QtCore.SIGNAL("clicked()"),
                               mainWindow.render_everything)
        QtCore.QMetaObject.connectSlotsByName(mainWindow)

    def retranslateUi(self, mainWindow):
        mainWindow.setWindowTitle(
            QtWidgets.QApplication.translate("mainWindow", "Q3D Endcaps", None,
                                             -1))
        self.label.setText(
            QtWidgets.QApplication.translate(
                "mainWindow", "Select endcap type for unconnected pins:", None,
                -1))
        self.renderButton.setText(
            QtWidgets.QApplication.translate("mainWindow", "Render", None, -1))



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""This code calculations the wavefunction(s) of the simple harmonic oscillator
corresponding to an LC circuit.

Key References:
    - *R. Shankar*, "Principles of Quantum Mechanics", Second Edition, Springer (1994)
    - or any other undergraduate quantum mechanics textbook)
"""

import matplotlib.pyplot as plt
import numpy as np
from math import *

__all__ = ['wavefunction']


def wavefunction(L, C, n, x):
    """This function calculates the nth wavefunction of the harmonic oscillator
    for a given value of inductance (L) and capacitance (C) at a charge x.

    Args:
        L (float): The inductance of the inductor in an LC circuit.
        C (float): The capacitance of the capacitor in an LC circuit.
        n (int): The energy state of the harmonic oscillator.
        x (float): The value of charge (independent variable) for which the wavefunction is calculated.

    Returns:
        float: Calculated wavefunction
    """
    # This is the fundamental frequency of the LC circuit
    omega = (L * C)**(0.5)

    # For simplicity, let's work in units where hbar=1
    hbar = 1.0

    # these are two terms in the expression for wavefunction which do not depend on charge (x)
    prefactor = np.sqrt(1.0 / ((2.0**n) * (np.math.factorial(n))))
    qubic_root = ((L * omega) / (np.pi * hbar))**(0.25)

    # these are the terms in the expression for the wavefunction which depend on charge (x) and energy state (n)
    if n == 0:
        Hermitian_term = prefactor * qubic_root * (2.71828**(
            (-L * omega * x**2.0) / (2.0 * hbar))) * 1.0
    if n == 1:
        Hermitian_term = prefactor * qubic_root * (2.71828**(
            (-L * omega * x**2.0) / (2.0 * hbar))) * (2.0 * x)
    if n == 2:
        Hermitian_term = prefactor * qubic_root * (2.71828**(
            (-L * omega * x**2.0) / (2.0 * hbar))) * (4.0 * x**2.0 - 2.0)
    if n == 3:
        Hermitian_term = prefactor * qubic_root * (2.71828**(
            (-L * omega * x**2.0) / (2.0 * hbar))) * (8.0 * x**3.0 - 12.0 * x)
    if n == 4:
        Hermitian_term = prefactor * qubic_root * (2.71828**(
            (-L * omega * x**2.0) /
            (2.0 * hbar))) * (16.0 * x**4.0 - 48.0 * x**2.0 + 12.0)

    # the final wavefunction is the product of these terms
    return (prefactor) * (qubic_root) * (Hermitian_term)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Transmon Pocket.

.. code-block::
     ________________________________
    |______ ____           __________|
    |      |____|         |____|     |
    |        __________________      |
    |       |                  |     |
    |       |__________________|     |
    |                 |              |
    |                 x              |
    |        _________|________      |
    |       |                  |     |
    |       |__________________|     |
    |        ______                  |
    |_______|______|                 |
    |________________________________|
"""

import numpy as np
from qiskit_metal import draw, Dict
from qiskit_metal.qlibrary.core import BaseQubit


class TransmonPocket(BaseQubit):
    """The base `TransmonPocket` class.

    Inherits `BaseQubit` class.

    Create a standard pocket transmon qubit for a ground plane,
    with two pads connected by a junction (see drawing below).

    Connector lines can be added using the `connection_pads`
    dictionary. Each connector pad has a name and a list of default
    properties.

    Sketch:
        Below is a sketch of the qubit
        ::

                 +1                            +1
                ________________________________
            -1  |______ ____           __________|   +1     Y
                |      |____|         |____|     |          ^
                |        __________________      |          |
                |       |     island       |     |          |----->  X
                |       |__________________|     |
                |                 |              |
                |  pocket         x              |
                |        _________|________      |
                |       |                  |     |
                |       |__________________|     |
                |        ______                  |
                |_______|______|                 |
            -1  |________________________________|   +1
                 
                 -1                            -1

    .. image::
        TransmonPocket.png

    .. meta::
        :description: Transmon Pocket

    BaseQubit Default Options:
        * connection_pads: Empty Dict -- The dictionary which contains all active connection lines for the qubit.
        * _default_connection_pads: Empty Dict -- The default values for the (if any) connection lines of the qubit.

    Default Options:
        * pad_gap: '30um' -- The distance between the two charge islands, which is also the resulting 'length' of the pseudo junction
        * inductor_width: '20um' -- Width of the pseudo junction between the two charge islands (if in doubt, make the same as pad_gap). Really just for simulating in HFSS / other EM software
        * pad_width: '455um' -- The width (x-axis) of the charge island pads
        * pad_height: '90um' -- The size (y-axis) of the charge island pads
        * pocket_width: '650um' -- Size of the pocket (cut out in ground) along x-axis
        * pocket_height: '650um' -- Size of the pocket (cut out in ground) along y-axis
        * _default_connection_pads: Dict
            * pad_gap: '15um' -- Space between the connector pad and the charge island it is nearest to
            * pad_width: '125um' -- Width (x-axis) of the connector pad
            * pad_height: '30um' -- Height (y-axis) of the connector pad
            * pad_cpw_shift: '5um' -- Shift the connector pad cpw line by this much away from qubit
            * pad_cpw_extent: '25um' -- Shift the connector pad cpw line by this much away from qubit
            * cpw_width: 'cpw_width' -- Center trace width of the CPW line
            * cpw_gap: 'cpw_gap' -- Dielectric gap width of the CPW line
            * cpw_extend: '100um' -- Depth the connector line extense into ground (past the pocket edge)
            * pocket_extent: '5um' -- How deep into the pocket should we penetrate with the cpw connector (into the fround plane)
            * pocket_rise: '65um' -- How far up or downrelative to the center of the transmon should we elevate the cpw connection point on the ground plane
            * loc_W: '+1' -- Width location  only +-1
            * loc_H: '+1' -- Height location only +-1
    """

    default_options = Dict(
        pad_gap='30um',
        inductor_width='20um',
        pad_width='455um',
        pad_height='90um',
        pocket_width='650um',
        pocket_height='650um',
        # 90 has dipole aligned along the +X axis,
        # while 0 has dipole aligned along the +Y axis
        _default_connection_pads=Dict(
            pad_gap='15um',
            pad_width='125um',
            pad_height='30um',
            pad_cpw_shift='5um',
            pad_cpw_extent='25um',
            cpw_width='cpw_width',
            cpw_gap='cpw_gap',
            # : cpw_extend: how far into the ground to extend the CPW line from the coupling pads
            cpw_extend='100um',
            pocket_extent='5um',
            pocket_rise='65um',
            loc_W='+1',  # width location  only +-1
            loc_H='+1',  # height location only +-1
        ))
    """Default drawing options"""

    component_metadata = Dict(short_name='Pocket',
                              _qgeometry_table_path='True',
                              _qgeometry_table_poly='True',
                              _qgeometry_table_junction='True')
    """Component metadata"""

    TOOLTIP = """The base `TransmonPocket` class."""

    def make(self):
        """Define the way the options are turned into QGeometry.

        The make function implements the logic that creates the geometry
        (poly, path, etc.) from the qcomponent.options dictionary of
        parameters, and the adds them to the design, using
        qcomponent.add_qgeometry(...), adding in extra needed
        information, such as layer, subtract, etc.
        """
        self.make_pocket()
        self.make_connection_pads()

    def make_pocket(self):
        """Makes standard transmon in a pocket."""

        # self.p allows us to directly access parsed values (string -> numbers) from the user option
        p = self.p

        # extract chip name
        chip = p.chip

        # since we will reuse these options, parse them once and define them as varaibles
        pad_width = p.pad_width
        pad_height = p.pad_height
        pad_gap = p.pad_gap

        # make the pads as rectangles (shapely polygons)
        pad = draw.rectangle(pad_width, pad_height)
        pad_top = draw.translate(pad, 0, +(pad_height + pad_gap) / 2.)
        pad_bot = draw.translate(pad, 0, -(pad_height + pad_gap) / 2.)

        rect_jj = draw.LineString([(0, -pad_gap / 2), (0, +pad_gap / 2)])
        # the draw.rectangle representing the josephson junction
        # rect_jj = draw.rectangle(p.inductor_width, pad_gap)

        rect_pk = draw.rectangle(p.pocket_width, p.pocket_height)

        # Rotate and translate all qgeometry as needed.
        # Done with utility functions in Metal 'draw_utility' for easy rotation/translation
        # NOTE: Should modify so rotate/translate accepts qgeometry, would allow for
        # smoother implementation.
        polys = [rect_jj, pad_top, pad_bot, rect_pk]
        polys = draw.rotate(polys, p.orientation, origin=(0, 0))
        polys = draw.translate(polys, p.pos_x, p.pos_y)
        [rect_jj, pad_top, pad_bot, rect_pk] = polys

        # Use the geometry to create Metal qgeometry
        self.add_qgeometry('poly',
                           dict(pad_top=pad_top, pad_bot=pad_bot),
                           chip=chip)
        self.add_qgeometry('poly',
                           dict(rect_pk=rect_pk),
                           subtract=True,
                           chip=chip)
        # self.add_qgeometry('poly', dict(
        #     rect_jj=rect_jj), helper=True)
        self.add_qgeometry('junction',
                           dict(rect_jj=rect_jj),
                           width=p.inductor_width,
                           chip=chip)

    def make_connection_pads(self):
        """Goes through connector pads and makes each one."""
        for name in self.options.connection_pads:
            self.make_connection_pad(name)

    def make_connection_pad(self, name: str):
        """Makes an individual connector.

        Args:
            name (str) : Name of the connector
        """

        # self.p allows us to directly access parsed values (string -> numbers) from the user option
        p = self.p
        pc = self.p.connection_pads[name]  # parser on connector options

        # extract chip name
        chip = p.chip

        # define commonly used variables once
        cpw_width = pc.cpw_width
        cpw_extend = pc.cpw_extend
        pad_width = pc.pad_width
        pad_height = pc.pad_height
        pad_cpw_shift = pc.pad_cpw_shift
        pocket_rise = pc.pocket_rise
        pocket_extent = pc.pocket_extent

        # Define the geometry
        # Connector pad
        connector_pad = draw.rectangle(pad_width, pad_height, -pad_width / 2,
                                       pad_height / 2)
        # Connector CPW wire
        connector_wire_path = draw.wkt.loads(f"""LINESTRING (\
            0 {pad_cpw_shift+cpw_width/2}, \
            {pc.pad_cpw_extent}                           {pad_cpw_shift+cpw_width/2}, \
            {(p.pocket_width-p.pad_width)/2-pocket_extent} {pad_cpw_shift+cpw_width/2+pocket_rise}, \
            {(p.pocket_width-p.pad_width)/2+cpw_extend}    {pad_cpw_shift+cpw_width/2+pocket_rise}\
                                        )""")
        # for connector cludge
        connector_wire_CON = draw.buffer(connector_wire_path, cpw_width /
                                         2.)  # helper for the moment

        # Position the connector, rot and tranlate
        loc_W, loc_H = float(pc.loc_W), float(pc.loc_H)
        if float(loc_W) not in [-1., +1.] or float(loc_H) not in [-1., +1.]:
            self.logger.info(
                'Warning: Did you mean to define a transmon wubit with loc_W and'
                ' loc_H that are not +1 or -1?? Are you sure you want to do this?'
            )
        objects = [connector_pad, connector_wire_path, connector_wire_CON]
        objects = draw.scale(objects, loc_W, loc_H, origin=(0, 0))
        objects = draw.translate(
            objects,
            loc_W * (p.pad_width) / 2.,
            loc_H * (p.pad_height + p.pad_gap / 2 + pc.pad_gap))
        objects = draw.rotate_position(objects, p.orientation,
                                       [p.pos_x, p.pos_y])
        [connector_pad, connector_wire_path, connector_wire_CON] = objects

        self.add_qgeometry('poly', {f'{name}_connector_pad': connector_pad},
                           chip=chip)
        self.add_qgeometry('path', {f'{name}_wire': connector_wire_path},
                           width=cpw_width,
                           chip=chip)
        self.add_qgeometry('path', {f'{name}_wire_sub': connector_wire_path},
                           width=cpw_width + 2 * pc.cpw_gap,
                           subtract=True,
                           chip=chip)

        ############################################################

        # add pins
        points = np.array(connector_wire_path.coords)
        self.add_pin(name,
                     points=points[-2:],
                     width=cpw_width,
                     input_as_norm=True,
                     chip=chip)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Transmon Pocket CL.

Pocket "axis"
        _________________
        |               |
        |_______________|       ^
        ________x________       |  N
        |               |       |
        |_______________|

Child of 'standard' transmon pocket.
"""
# pylint: disable=invalid-name
# Modification of Transmon Pocket Object to include a charge line (would be better to just make as a child)

import numpy as np
from qiskit_metal import draw, Dict
from qiskit_metal.qlibrary.qubits.transmon_pocket import TransmonPocket


class TransmonPocketCL(TransmonPocket):  # pylint: disable=invalid-name
    """The base `TransmonPocketCL` class.

    Inherits `TransmonPocket` class.

    Create a standard pocket transmon qubit for a ground plane,
    with two pads connected by a junction (see drawing below).

    Connector lines can be added using the `connection_pads`
    dictionary. Each connector line has a name and a list of default
    properties.

    This is a child of TransmonPocket, see TransmonPocket for the variables and
    description of that class.

    ::

        _________________
        |               |
        |_______________|       ^
        ________x________       |  N
        |               |       |
        |_______________|


    .. image::
        TransmonPocketCL.png

    .. meta::
        :description: Transmon Pocket Charge Line

    BaseQubit Default Options:
        * connection_pads: empty Dict -- The dictionary which contains all active connection lines for the qubit.
        * _default_connection_pads: empty Dict -- The default values for the (if any) connection lines of the qubit.

    TransmonPocket Default Options:
        * pad_gap: '30um' -- The distance between the two charge islands, which is also the resulting 'length' of the pseudo junction
        * inductor_width: '20um' -- Width of the pseudo junction between the two charge islands (if in doubt, make the same as pad_gap). Really just for simulating in HFSS / other EM software
        * pad_width: '455um' -- The width (x-axis) of the charge island pads
        * pad_height: '90um' -- The size (y-axis) of the charge island pads
        * pocket_width: '650um' -- Size of the pocket (cut out in ground) along x-axis
        * pocket_height: '650um' -- Size of the pocket (cut out in ground) along y-axis
        * _default_connection_pads: Dict
            * pad_gap: '15um' -- Space between the connector pad and the charge island it is nearest to
            * pad_width: '125um' -- Width (x-axis) of the connector pad
            * pad_height: '30um' -- Height (y-axis) of the connector pad
            * pad_cpw_shift: '5um' -- Shift the connector pad cpw line by this much away from qubit
            * pad_cpw_extent: '25um' -- Shift the connector pad cpw line by this much away from qubit
            * cpw_width: 'cpw_width' -- Center trace width of the CPW line
            * cpw_gap: 'cpw_gap' -- Dielectric gap width of the CPW line
            * cpw_extend: '100um' -- Depth the connector line extends into ground (past the pocket edge)
            * pocket_extent: '5um' -- How deep into the pocket should we penetrate with the cpw connector (into the fround plane)
            * pocket_rise: '65um' -- How far up or downrelative to the center of the transmon should we elevate the cpw connection point on the ground plane
            * loc_W: '+1' -- Width location  only +-1
            * loc_H: '+1' -- Height location only +-1

    Default Options:
        * make_CL: True -- If a charge line should be included.
        * cl_gap: '6um' -- The cpw dielectric gap of the charge line
        * cl_width: '10um' -- The cpw trace width of the charge line
        * cl_length: '20um' --  The length of the charge line 'arm' coupling the the qubit pocket.
          Measured from the base of the 90 degree bend.
        * cl_ground_gap: '6um' -- How much ground is present between the charge line and the
          qubit pocket.
        * cl_pocket_edge: '0' -- What side of the pocket the charge line is.
          -180 to +180 from the 'west edge', will round to the nearest 90.
        * cl_off_center: '50um' -- Distance from the center axis the qubit pocket is referenced to
    """
    component_metadata = Dict(short_name='Q', _qgeometry_table_poly='True')
    """Component metadata"""

    default_options = Dict(
        make_CL=True,
        cl_gap='6um',  # the cpw dielectric gap of the charge line
        cl_width='10um',  # the cpw trace width of the charge line
        # the length of the charge line 'arm' coupling the the qubit pocket.
        cl_length='20um',
        # Measured from the base of the 90 degree bend
        cl_ground_gap=
        '6um',  # how much ground between the charge line and the qubit pocket
        # -180 to +180 from the 'left edge', will round to the nearest 90.
        cl_pocket_edge='0',
        cl_off_center=
        '50um',  # distance from the center axis the qubit pocket is built on
    )
    """Default drawing options"""

    TOOLTIP = """Create a standard pocket transmon qubit for a ground plane,
    with two pads connected by a junction"""

    def make(self):
        """Define the way the options are turned into QGeometry."""
        super().make()

        if self.options.make_CL == True:
            self.make_charge_line()


#####################################################################

    def make_charge_line(self):
        """Creates the charge line if the user has charge line option to
        TRUE."""

        # Grab option values
        name = 'Charge_Line'

        p = self.p

        cl_arm = draw.box(0, 0, -p.cl_width, p.cl_length)
        cl_cpw = draw.box(0, 0, -8 * p.cl_width, p.cl_width)
        cl_metal = draw.unary_union([cl_arm, cl_cpw])

        cl_etcher = draw.buffer(cl_metal, p.cl_gap)

        port_line = draw.LineString([(-8 * p.cl_width, 0),
                                     (-8 * p.cl_width, p.cl_width)])

        polys = [cl_metal, cl_etcher, port_line]

        # Move the charge line to the side user requested
        cl_rotate = 0
        if (abs(p.cl_pocket_edge) > 135) or (abs(p.cl_pocket_edge) < 45):
            polys = draw.translate(
                polys, -(p.pocket_width / 2 + p.cl_ground_gap + p.cl_gap),
                p.cl_off_center)
            if (abs(p.cl_pocket_edge) > 135):
                cl_rotate = 180
        else:
            polys = draw.translate(
                polys, -(p.pocket_height / 2 + p.cl_ground_gap + p.cl_gap),
                p.cl_off_center)
            cl_rotate = 90
            if (p.cl_pocket_edge < 0):
                cl_rotate = -90

        # Rotate it to the pockets orientation
        polys = draw.rotate(polys, p.orientation + cl_rotate, origin=(0, 0))

        # Move to the final position
        polys = draw.translate(polys, p.pos_x, p.pos_y)

        [cl_metal, cl_etcher, port_line] = polys

        # Generating pins
        points = list(draw.shapely.geometry.shape(port_line).coords)
        self.add_pin(name, points, p.cl_width)

        # Adding to qgeometry table
        self.add_qgeometry('poly', dict(cl_metal=cl_metal))
        self.add_qgeometry('poly', dict(cl_etcher=cl_etcher), subtract=True)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Main module for shapely manipulation and drawing.

**Manipulation:**
    The module defines rotate, scale, translate to operate on a various set of
    input arguments, such as shapelies or itterable/mappables or shapelies.

**Creation:**
    The module defines the creation of some basic shapely objects for convinence,
    such as a rectangle.
"""

from collections.abc import Iterable, Mapping

import numpy as np
import shapely
import shapely.affinity
import shapely.wkt
from shapely.geometry import CAP_STYLE, JOIN_STYLE, Point, Polygon
from shapely.geometry.multipolygon import MultiPolygon

from .. import config, logger
from .. import is_component
from ..config import DefaultMetalOptions
from . import BaseGeometry
from .utility import get_poly_pts

__all__ = [
    'rectangle', 'is_rectangle', 'flip_merge', 'rotate', 'rotate_position',
    '_iter_func_geom_', 'translate', 'scale', 'buffer', 'union', 'subtract'
]


def rectangle(w: float, h: float, xoff: float = 0, yoff: float = 0):
    """Draw a shapely rectangle of width and height.

    Args:
        w (float) :  width
        h (float) :  height
        xoff (float, optional) : Gives the x position of the center.  Defaults to 0.
        yoff (float, optional) : Gives the y position of the center.  Defaults to 0.

    Returns:
        shapely.geometry.Polygon: Shapely rectangle
    """

    w, h = float(w), float(h)
    pad = f"""POLYGON (({-w/2} {-h/2},
                        {+w/2} {-h/2},
                        {+w/2} {+h/2},
                        {-w/2} {+h/2},
                        {-w/2} {-h/2}))"""  # last  point has to close on self

    if xoff is 0 and yoff is 0:
        return Polygon(shapely.wkt.loads(pad))  # My Polygon class
    else:
        return shapely.affinity.translate(Polygon(shapely.wkt.loads(pad)),
                                          xoff=xoff,
                                          yoff=yoff)


def is_rectangle(obj):
    """Test if a shapely object is a rectangle.

    If there are 4 ext cooridnate then
    check if consecutive vectors are orthogonal.
    Assumes that the last point is not repeating.

    Args:
        obj (object): Object to test

    Returns:
        bool: True is object is a rectangle, false otherwise
    """
    if not isinstance(obj, shapely.geometry.Polygon):
        return False

    p = get_poly_pts(obj)
    if len(p) == 4:

        def is_orthogonal(i):
            v1 = p[(i + 1) % 4] - p[(i + 0) % 4]
            v2 = p[(i + 2) % 4] - p[(i + 1) % 4]
            return abs(np.dot(v1, v2)) < 1E-16

        # CHeck if all vectors are consequtivly orthogonal
        return all(map(is_orthogonal, range(4)))
    else:
        return False


def subtract(poly_main: shapely.geometry.Polygon,
             poly_tool: shapely.geometry.Polygon):
    """Geometry subtract tool poly from main poly.

    Args:
        poly_main (Polygon): Main polygon from which we will carve out
        poly_tool (Polygon or list): Poly or list of polys to subtract

    Returns:
        Polygon: Difference between the given Polygons
    """
    return poly_main.difference(poly_tool)


def union(*polys):
    """Geometry union of two or more polys.

    Areas of overlapping Polygons will get merged. LineStrings will get
    fully dissolved and noded. Duplicate Points will get merged.

    Args:
        polys (Polygon, or list of polygons, or args of polygons).
            Main polygon from which we will carve out.

    Returns:
        Polygon: Union of the input polygons
    """
    # union() is an expensive way to find the cumulative union of
    # many objects.
    # See shapely.ops.unary_union() for a more effective method.
    if len(polys) is 2:
        return polys[0].union(polys[1])
    elif len(polys) is 1:  # assume that it is a list
        polys = polys[0]
    return shapely.ops.unary_union(polys)


#########################################################################
# Shapely affine transorms


def flip_merge(line: shapely.geometry.LineString,
               xfact=-1,
               yfact=+1,
               origin=(0, 0)):
    """Mirror type of drawing command on LineString.

    Calls shapely.affinity.scale(geom, xfact=xfact, yfact=yfact, zfact=1.0, origin=origin).

    The point of origin can be a keyword 'center' for the 2D bounding
    box center (default), 'centroid' for the geometry's 2D centroid,
    a Point object or a coordinate tuple (x0, y0, z0).
    Negative scale factors will mirror or reflect coordinates.

    Args:
        line (shapely.geometry.LineString): Geometry
        xfact (double): x-scale factor.  Defaults to -1.
        yfact (double): y-scale factor.  Defaults to 1.
        orign (tuple): Coordinates of the origin

    Returns:
        list: The coorindates that can be used to construct a polygon, by flipping a linestring
        over an axis and mirroring it.
    """
    line_flip = shapely.affinity.scale(line,
                                       xfact=xfact,
                                       yfact=yfact,
                                       origin=origin)
    coords = list(line.coords) + list(reversed(line_flip.coords))
    return coords


def _iter_func_geom_(func, objs, *args, overwrite=False, **kwargs):
    """Apply function to shapely geometry, and handle. This will create a new
    dictionary.

    Applied to on:
        For components:
            objs.qgeometry
        For qgeometry:
            qgeometry.geom
            # not for now: qgeometry.geom_rendred.metal
        For dict:
            to each value
        For list:
            to each element

    Args:
        func (function): Function the apply
        objs (Dict, List, Tuple or BaseGeometry): Set of objects
        overwrite (bool): Overwrite the parent dict or not. This applies to component qgeometry dictionary.
            Maybe remove in future?  Defaults to False.
        kwargs (dict): Parameters dictionary

    Returns:
        list: List of objects
    """

    if isinstance(objs, Mapping):  # dict, Dict
        if overwrite:
            for key, val in objs.items():
                objs[key] = _iter_func_geom_(func,
                                             val,
                                             *args,
                                             overwrite=overwrite,
                                             **kwargs)
            return objs
        else:
            return type(objs)([(key,
                                _iter_func_geom_(func,
                                                 val,
                                                 *args,
                                                 overwrite=overwrite,
                                                 **kwargs))
                               for key, val in objs.items()])

    elif isinstance(objs, Iterable):  # list, tuple
        if overwrite:
            for key, val in enumerate(objs):
                objs[key] = _iter_func_geom_(func,
                                             val,
                                             *args,
                                             overwrite=overwrite,
                                             **kwargs)
            return objs
        # Under Shapely 1.8, MultiPolygon objects need to be
        # referenced with the .geoms property. This function
        # did inadvertently take MultiPolygon objects so
        # another case was added to handle this.
        elif isinstance(objs, MultiPolygon):
            return type(objs)([
                _iter_func_geom_(func,
                                 val,
                                 *args,
                                 overwrite=overwrite,
                                 **kwargs) for val in objs.geoms
            ])
        else:
            return type(objs)([
                _iter_func_geom_(func,
                                 val,
                                 *args,
                                 overwrite=overwrite,
                                 **kwargs) for val in objs
            ])

    elif is_component(objs):
        # apply on geom of component's qgeometry; return component
        # below: returns a dict with the shifted components, we will not use this
        _iter_func_geom_(func,
                         objs.qgeometry,
                         *args,
                         overwrite=overwrite,
                         **kwargs)
        return objs

    # elif is_element(objs): #will need updating to new table format
    #    # apply on geom of element; return element
    #    func(objs.geom, *args, **kwargs)
    #    return objs

    elif isinstance(objs, BaseGeometry):
        return func(objs, *args, **kwargs)

    else:
        logger.error(f'Unkown elemnet! ERROR on: {objs}')
        return objs


def rotate(qgeometry,
           angle,
           origin='center',
           use_radians=False,
           overwrite=False):
    r"""
    The angle of rotation can be specified in either degrees (default) or
    radians by setting ``use_radians=True``. Positive angles are
    counter-clockwise and negative are clockwise rotations.

    The point of origin can be a keyword 'center' for the bounding box
    center (default), 'centroid' for the geometry's centroid, a Point object
    or a coordinate tuple (x0, y0).

    Calls: shapely.affinity.rotate(
        geom, angle, origin='center', use_radians=False)

    Args:
        qgeometry (Dict, List, Tuple or BaseGeometry): Set of objects
        angle (double): Rotation angle
        origin (tuple or str): Origin point. Defaults to 'center'.
        use_radians (bool): True to use radians.  Defaults to False.
        overwrite (bool): True to overwrite.  Defaults to False.

    Returns:
        geometry: A rotated geometry on a 2D plane.

    The affine transformation matrix for 2D rotation is:
    ::

        / cos(r) -sin(r) xoff \
        | sin(r)  cos(r) yoff |
        \   0       0      1  /

    where the offsets are calculated from the origin Point(x0, y0):
    ::

        xoff = x0 - x0 * cos(r) + y0 * sin(r)
        yoff = y0 - x0 * sin(r) - y0 * cos(r)
    """
    return _iter_func_geom_(shapely.affinity.rotate,
                            qgeometry,
                            angle,
                            origin=origin,
                            use_radians=use_radians,
                            overwrite=overwrite)


def translate(qgeometry, xoff=0.0, yoff=0.0, zoff=0.0, overwrite=False):
    r'''Shifts the geometry by the given offset.
    Args:
        qgeometry (Dict, List, Tuple or BaseGeometry): Set of objects.
        xoff (double): x-direction offset.  Defaults to 0.0.
        yoff (double): y-direction offset.  Defaults to 0.0.
        zoff (double): z-direction offset.  Defaults to 0.0.
        overwrite (bool): True to overwrite.  Defaults to False.

    Returns:
        geometry: A translated geometry shifted by offsets along each dimension.

    The general 3D affine transformation matrix for translation is:
    ::

        / 1  0  0 xoff \
        | 0  1  0 yoff |
        | 0  0  1 zoff |
        \ 0  0  0   1  /
    '''
    return _iter_func_geom_(shapely.affinity.translate,
                            qgeometry,
                            xoff=xoff,
                            yoff=yoff,
                            zoff=zoff,
                            overwrite=overwrite)


def scale(qgeometry,
          xfact=1.0,
          yfact=1.0,
          zfact=1.0,
          origin='center',
          overwrite=False):
    r'''
    Operatos on a list or Dict of components.

    Args:
        qgeometry (Dict, List, Tuple or BaseGeometry): Set of objects
        xfact: x-direction scale factor.  Defaults to 1.0.
        yfact: y-direction scale factor.  Defaults to 1.0.
        zfact: z-direction scale factor.  Defaults to 1.0.
        origin (tuple or str): Origin point.  Defaults to 'center'.
        overwrite (bool): True to overwrite.  Defaults to False.

    Returns:
        geometry: A scaled geometry, scaled by factors along each dimension.

    The point of origin can be a keyword 'center' for the 2D bounding box
    center (default), 'centroid' for the geometry's 2D centroid, a Point
    object or a coordinate tuple (x0, y0, z0).

    Negative scale factors will mirror or reflect coordinates.

    The general 3D affine transformation matrix for scaling is:
    ::

        / xfact  0    0   xoff \
        |   0  yfact  0   yoff |
        |   0    0  zfact zoff |
        \   0    0    0     1  /

    where the offsets are calculated from the origin Point(x0, y0, z0):
    ::

        xoff = x0 - x0 * xfact
        yoff = y0 - y0 * yfact
        zoff = z0 - z0 * zfact
    '''
    return _iter_func_geom_(shapely.affinity.scale,
                            qgeometry,
                            xfact=xfact,
                            yfact=yfact,
                            zfact=zfact,
                            origin=origin,
                            overwrite=overwrite)


def rotate_position(qgeometry,
                    angle: float,
                    pos: list,
                    pos_rot=(0, 0),
                    overwrite=False):
    """Orient and then place position. Just a shortcut function.

    Args:
        qgeometry (Dict, List, Tuple or BaseGeometry): Set of objects
        angle (float): Rotation angle
        pos (list, np.array): Position to translate to
        pos_rot (tuple): Rotate about this point before translating.  Default to {(0, 0)}).

    Returns:
        geometry: Rotate dand translated, same as input
    """

    def rotate_position_shapely(sobj):
        pos1 = list(shapely.affinity.rotate(Point(pos), angle).coords)[0]
        sobj = shapely.affinity.rotate(sobj, angle,
                                       pos_rot)  # rotate about pos_rot
        return shapely.affinity.translate(sobj, *pos1)  # move to position

    return _iter_func_geom_(rotate_position_shapely,
                            qgeometry,
                            overwrite=overwrite)


def buffer(qgeometry,
           distance: float,
           resolution=None,
           cap_style=CAP_STYLE.flat,
           join_style=JOIN_STYLE.mitre,
           mitre_limit=None,
           overwrite=False):
    """Flat buffer of all components in the dictionary.

    Default stlye:
        - cap_style=CAP_STYLE.flat
        - join_style=JOIN_STYLE.mitre

    Args:
        qgeometry (Dict, List, Tuple or BaseGeometry): Set of objects.
        distance (float): Distance.
        resolution (int): How many points.  Defaults to None.
        cap_style (shapely.geometry.CAP_STYLE): Cap style.  Defaults to CAP_STYLE.falt.
        join_style (shapely.geometry.JOIN_STYLE): Join style.  Defaults to JOIN_STYLE.mitre.
        mitre_limit (double): Mitre limit.  Defaults to None.
        overwrite (bool): True to overwrite.  Defaults to None.

    Returns:
        geometry: A geometry with an envelope at a distance from the object's envelope

    A negative distance has a "shrink" effect. A zero distance may be used
    to "tidy" a polygon. The resolution of the buffer around each vertex of
    the object increases by increasing the resolution keyword parameter
    or second positional parameter. resolution = how many points

    The styles of caps are: CAP_STYLE.round (1), CAP_STYLE.flat (2), and
    CAP_STYLE.square (3).

    The styles of joins between offset segments are: JOIN_STYLE.round (1),
    JOIN_STYLE.mitre (2), and JOIN_STYLE.bevel (3).

    The mitre limit ratio is used for very sharp corners. The mitre ratio
    is the ratio of the distance from the corner to the end of the mitred
    offset corner. When two line segments meet at a sharp angle, a miter
    join will extend the original geometry. To prevent unreasonable
    geometry, the mitre limit allows controlling the maximum length of the
    join corner. Corners with a ratio which exceed the limit will be
    beveled.

    Example use:
    ::

        x = rectangle(1,1)
        y = buffer_flat([x,x,[x,x,{'a':x}]], 0.5)
        draw.mpl.render([x,y])
        units = config.DefaultMetalOptions.default_generic.units
    """
    if mitre_limit is None:
        mitre_limit = DefaultMetalOptions.default_generic.geometry.buffer_mitre_limit
        # TODO: maybe this should be in the renderer for metal?
        # or maybe render can set config? - yes
        # TODO: should really be design.template_options.units

    if resolution is None:
        resolution = DefaultMetalOptions.default_generic.geometry.buffer_resolution

    def buffer_me(obj, *args, **kwargs):
        return obj.buffer(*args, **kwargs)

    return _iter_func_geom_(buffer_me,
                            qgeometry,
                            distance,
                            resolution=resolution,
                            cap_style=cap_style,
                            join_style=join_style,
                            mitre_limit=mitre_limit,
                            overwrite=overwrite)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=protected-access
# pylint: disable=global-statement
# pylint: disable=import-error
"""Math and override functions."""

import numpy as np

__all__ = ['set_decimal_precision', 'dot', 'cross', 'round']

DECIMAL_PRECISION = 10


def set_decimal_precision(value: int):
    """Override the decimal_precision default (10).

    Args:
        value: Any integer. If present, decimal part will be truncated (flooring)
    """
    global DECIMAL_PRECISION
    DECIMAL_PRECISION = int(value)


def dot(vector_1: np.array, vector_2: np.array) -> float:
    """Numpy dot product with decimal_precision.

    Args:
        vector_1 (np.array): First of the dot product vectors
        vector_2 (np.array): Second of the dot product vectors

    Returns:
        float: Rounded dot product
    """
    return round(np.dot(vector_1, vector_2))


# pylint: disable=redefined-builtin
def round(value) -> float:
    """Numpy rounding with decimal_precision.

    Args:
        value: Any numerical type supported by np.round()

    Returns:
        float: Rounded number
    """
    return np.round(value, DECIMAL_PRECISION)


def cross(vector_1: np.array, vector_2: np.array) -> float:
    """Numpy cross product with decimal_precision.

    Args:
        vector_1 (np.array): First of the cross product vectors
        vector_2 (np.array): Second of the cross product vectors

    Returns:
        float: Rounded cross product
    """
    return round(np.cross(vector_1, vector_2))


def aligned_pts(points: list) -> bool:
    """Are the three points aligned? with decimal_precision.

    Args:
        points (list): of 3 points expressed as list or np.array

    Returns:
        bool: True if they are aligned. False otherwise
    """
    if len(points) != 3:
        raise Exception(
            "Ambiguous. You can only pass exactly 3 points to this method")
    v1 = points[1] - points[0]
    v1_dir = v1 / np.linalg.norm(v1)
    v2 = points[2] - points[1]
    v2_dir = v2 / np.linalg.norm(v2)
    result = dot(v1_dir, v2_dir)
    return True if result == 1 else False
